package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.swing.JOptionPane;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EParameter;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.EcorePackage;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.presentation.EcoreEditor;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.ui.handlers.HandlerUtil;



public class AddEParameterConnector extends RefactoringConnector {
	

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		activeEditor = HandlerUtil.getActiveEditor(event);
		window = HandlerUtil.getActiveWorkbenchWindow(event);

		if (activeEditor instanceof EcoreEditor) {
			EcoreEditor ecoreEditor = (EcoreEditor) activeEditor;
			ecoreEditor.doSave(null);
			TreeSelection sel = (TreeSelection) ecoreEditor.getSelection();
			Object obj = sel.getFirstElement();

			if (obj instanceof EOperation) {
				EOperation op = (EOperation) obj;
				EPackage pack = op.getEContainingClass().getEPackage();
				List<String> classifierNames = new LinkedList<String>();
				List<EClassifier> classes = pack.getEClassifiers();
				for (EClassifier c : classes) {
					classifierNames.add(c.getName());
				}
				classifierNames.add("EInt");
				classifierNames.add("EDouble");
				classifierNames.add("EBoolean");
				classifierNames.add("EChar");
				classifierNames.add("EString");
				
				ChangeRecorder recorder = new ChangeRecorder();
				Map<EObject, URI> eObjectToProxyURIMap = new HashMap<EObject, URI>();
				recorder.setEObjectToProxyURIMap(eObjectToProxyURIMap);
				recorder.beginRecording(Collections.singleton(op));

				initRefactoring(op.getEContainingClass().getEPackage()
						.eContents());

				String paramName = JOptionPane
						.showInputDialog("Enter the name for the new parameter");
				String type = getSelectedFromDialog(classifierNames,
						"Set parameter type.");

				EClassifier paramType = EcoreFactory.eINSTANCE
						.createEDataType();
				if (type.equals("EInt")) {
					EDataType eDataType = EcoreFactory.eINSTANCE.createEDataType();
					eDataType = EcorePackage.Literals.EINT;
					paramType = eDataType;
				} else if (type.equals("EDouble")) {
					EDataType eDataType = EcoreFactory.eINSTANCE.createEDataType();
					eDataType = EcorePackage.Literals.EDOUBLE;
					paramType = eDataType;
				} else if (type.equals("EBoolean")) {
					EDataType eDataType = EcoreFactory.eINSTANCE.createEDataType();
					eDataType = EcorePackage.Literals.EBOOLEAN;
					paramType = eDataType;
				} else if (type.equals("EChar")) {
					EDataType eDataType = EcoreFactory.eINSTANCE.createEDataType();
					eDataType = EcorePackage.Literals.ECHAR;
					paramType = eDataType;
				} else if (type.equals("EString")) {
					EDataType eDataType = EcoreFactory.eINSTANCE.createEDataType();
					eDataType = EcorePackage.Literals.ESTRING;
					paramType = eDataType;
				}
				if (!type.equals("EInt") && !type.equals("EDouble")
						&& !type.equals("EBoolean") && !type.equals("EChar")
						&& !type.equals("EString")) {
					try {
						paramType = helper.getClass(type);
					} catch (Throwable e1) {
						// TODO Auto-generated catch block
						e1.printStackTrace();
					}
				}

				try {
					refactoring.addEParameter(op.getName(), paramName, paramType);					
				} catch (Throwable e) {
					JOptionPane
							.showMessageDialog(
									null,
									"The operation has already got a parameter with the same name.",
									"ERROR", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
					return null;
				}
				try {
					op.getEContainingClass().getEPackage().eResource().save(Collections.EMPTY_MAP);
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				
				ChangeDescription description = recorder.endRecording();
				if (!(description.eContents().isEmpty())) {
					propagateChanges(description, op, op.eResource());
				}
			}
		}
		return null;

	}

	public boolean checkPrecondition(EOperation op, String name) {
		EList<EParameter> params = op.getEParameters();
		for (EParameter p : params)
			if (p.getName().equals(name))
				return false;
		return true;
	}


	/**
	 * Changes OPERATION_CALLs according to the corresponding EOperation's parameter list.
	 */
	@Override
	protected void applyToAffectedObjects(ENamedElement eNamedElement) {
		EList<Object> temp = new BasicEList<Object>();
		for (Resource res : gtResources) {
			temp.add((Object) res);		
		}
		propagation.updateOperationCalls(temp);
		for (Resource res : gtResources) {
			propagation.cleanUpAndSave(null, res);
			
			URI resURI = res.getURI();
			String fragment = resURI.toString();
			String[] resRuleNameArray = fragment.split("/");
			String resRuleName = resRuleNameArray[resRuleNameArray.length-1];			
			int response = JOptionPane.showConfirmDialog(null,
					"Would you like to validate the GT rule " + resRuleName + "?", "Attention",
					JOptionPane.YES_NO_OPTION);
			if(response == 0){
				try{
					boolean result = callValidation(res);
					if(result == false){
						JOptionPane.showMessageDialog(null,
								"Errors exist in the rule " + resRuleName + "\nPlease check it manually.", "Validation failed!",
								JOptionPane.ERROR_MESSAGE);
					}
				} catch (Throwable e){
					e.printStackTrace();
				}
			}
		}
	}
}