package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.util.Collections;
import java.util.List;

import javax.swing.JOptionPane;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EParameter;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.presentation.EcoreEditor;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.gmf.runtime.notation.Edge;
import org.eclipse.gmf.runtime.notation.Node;
import org.eclipse.gmf.runtime.notation.Shape;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;

import de.ubt.ai1.modGraph.example.eRefactoring.ClassType;
import de.ubt.ai1.modgraph.gt.gtEngine.GTField;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPath;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPostcondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPrecondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GraphTransformationRule;

public class CollapseHierarchyConnector extends RefactoringConnector {
	private IEditorPart activeEditor;
	private EClass sourceClass;
	private EClass targetClass;

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		activeEditor = HandlerUtil.getActiveEditor(event);
		window = HandlerUtil.getActiveWorkbenchWindow(event);

		if (activeEditor instanceof EcoreEditor) {
			EcoreEditor ecoreEditor = (EcoreEditor) activeEditor;
			ecoreEditor.doSave(null);
			TreeSelection sel = (TreeSelection) ecoreEditor.getSelection();
			Object obj = sel.getFirstElement();

			if (obj instanceof EClass) {

				EClass subClass = (EClass) obj;
				// SW
				ChangeRecorder recorder = new ChangeRecorder(
						subClass.eContainer());

				Resource res = subClass.getEPackage().eResource();
				EClass superClass = null;
				if (!subClass.getESuperTypes().isEmpty())
					superClass = subClass.getESuperTypes().get(0);
				else {
					JOptionPane.showMessageDialog(null,
							"Select a subclass please.", "ERROR",
							JOptionPane.ERROR_MESSAGE);
					return null;
				}

				List<EObject> objects = subClass.getEPackage().eContents();
				initRefactoring(objects);
				ClassType classType = ClassType.SUPER_CLASS;
				String[] choices = { superClass.getName(), subClass.getName() };
				int response = JOptionPane.showOptionDialog(null,
						"Which class should be removed?", "Choose a class",
						JOptionPane.OK_CANCEL_OPTION,
						JOptionPane.PLAIN_MESSAGE, null, choices,
						"None of your business");

				switch (response) {
				case 0:
					classType = ClassType.SUPER_CLASS;
					sourceClass = superClass;
					targetClass = subClass;
					break;
				case 1:
					classType = ClassType.SUB_CLASS;
					sourceClass = subClass;
					targetClass = superClass;
					break;
				case -1:
					return null;
				default:
					JOptionPane.showMessageDialog(null, "Unexpected response "
							+ response);
				}

				for (EReference r : targetClass.getEReferences()) {
					if (r.getEType().equals(sourceClass))
						targetClass.getEStructuralFeatures().remove(r);
				}
				refactoring.collapseHierararchy(classType,
						superClass.getName(), subClass.getName());
				// inlined Class remains in the helper and needs to be
				// removed, so that updateOperationCalls() will work
				// properly
				EList<EObject> temp = new BasicEList<EObject>();
				for (EClass eClass : helper.getAllClasses())
					for (EOperation eOp : eClass.getEOperations())
						for (EParameter eParam : eOp.getEParameters()) {
							if (eParam.getEType().eIsProxy()
									|| eParam.getEType() == null)
								temp.add(eParam);
							else if (eParam.getEType().getName()
									.equals(sourceClass.getName())) {
								temp.add(eParam);
								temp.add(sourceClass);
							}
						}
				for (int i = temp.size() - 1; i >= 0; i--)
					EcoreUtil.delete(temp.get(i));
				
				try {
					res.save(Collections.EMPTY_MAP);
				} catch (Throwable e) {
					e.printStackTrace();
				}
				// SW
				ChangeDescription description = recorder.endRecording();
				if (!(description.eContents().isEmpty())) {
					propagateChanges(description, targetClass, targetClass.eResource());
				}

			}

		}
		return null;

	}

	@SuppressWarnings("rawtypes")
	@Override
	protected void applyToAffectedObjects(ENamedElement eNamedElement) {
		EList<Object> temp = new BasicEList<Object>();
		for (Resource res : gtResources) {
			temp.add((Object) res);
			EcoreUtil.resolveAll(res);
			TreeIterator<EObject> iter = res.getAllContents();
			EList<EObject> objectsToRemove = new BasicEList<EObject>();
			String ruleName = "";
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					ruleName = ((GraphTransformationRule) eObject).getName();
					objectsToRemove = propagation
							.propagateInlineClass_CollapseHierarchy(
									sourceClass, targetClass,
									(GraphTransformationRule) eObject,
									objectsToRemove);
				} 
				else if (eObject instanceof Node) {
					if (objectsToRemove.contains(((Node) eObject).getElement()))
						objectsToRemove.add(eObject);
				} 
				else if (eObject instanceof Edge) {
					if (objectsToRemove.contains(((Edge) eObject).getElement()))
						objectsToRemove.add(eObject);
				}
				else if (eObject instanceof GTField	&& ((GTField) eObject).getKind().getName().equals("OCL_CONSTRAINT")) {
					String containerName = "OCL_CONSTRAINT";
					String expression = ((GTField) eObject).getContent();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, sourceClass.getName(), containerName);
					if(newExpression != null)
						((GTField) eObject).setContent(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if (eObject instanceof GTPath) {
					String containerName = "GTPath";
					String expression = ((GTPath) eObject).getExpression();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, sourceClass.getName(), containerName);
					if(newExpression != null)
						((GTPath) eObject).setExpression(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPrecondition){
					String containerName = "GTPrecondition";
					String expression = ((GTPrecondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, sourceClass.getName(), containerName);
					if(newExpression != null)
						((GTPrecondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPostcondition){
					String containerName = "GTPostcondition";
					String expression = ((GTPostcondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, sourceClass.getName(), containerName);
					if(newExpression != null)
						((GTPostcondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof Shape && objectsToRemove.contains(((Shape)eObject).getElement()))
					objectsToRemove.add(eObject);
			}
			propagation.cleanUpAndSave(objectsToRemove, res);
		}
		propagation.updateOperationCalls(temp);
		for (Resource res : gtResources) {
			propagation.cleanUpAndSave(null, res);
			
			URI resURI = res.getURI();
			String fragment = resURI.toString();
			String[] resRuleNameArray = fragment.split("/");
			String resRuleName = resRuleNameArray[resRuleNameArray.length-1];			
			int response = JOptionPane.showConfirmDialog(null,
					"Would you like to validate the GT rule " + resRuleName + "?", "Attention",
					JOptionPane.YES_NO_OPTION);
			if(response == 0){
				try{
					boolean result = callValidation(res);
					if(result == false){
						JOptionPane.showMessageDialog(null,
								"Errors exist in the rule " + resRuleName + "\nPlease check it manually.", "Validation failed!",
								JOptionPane.ERROR_MESSAGE);
					}
				} catch (Throwable e){
					e.printStackTrace();
				}
			}
		}
	}

	// /**
	// * Moves attributes and fields. Reroutes or deletes links if necessary.
	// * Removes the corresponding node.
	// */
	// @SuppressWarnings({ "rawtypes", "unchecked" })
	// @Override
	// protected void applyToAffectedObjects(ENamedElement eNamedElement) {
	// for (Resource res : gtResources) {
	// TreeIterator<EObject> iter = res.getAllContents();
	// List<EObject> objectsToRemove = new ArrayList<EObject>();
	// while (iter.hasNext()) {
	// EObject eObject = iter.next();
	// EClass sourceClass;
	// String name = "";
	// // source node instanceof GPObject
	// if (eObject instanceof GPObject) {
	// sourceClass = (EClass) ((GPObject) eObject).getType();
	// if (sourceClass.eIsProxy()) {
	// URI classURI = EcoreUtil.getURI(sourceClass);
	// String fragment = classURI.fragment();
	// name = fragment.split("/")[fragment.split("/").length - 1];
	// } else {
	// name = sourceClass.getName();
	// }
	// if (name.equals(this.sourceClass.getName())) {
	// objectsToRemove.add(eObject);
	// GraphTransformationRule<EObject, EObject> rule =
	// (GraphTransformationRule<EObject, EObject>) ((GPObject) eObject)
	// .eContainer().eContainer();
	// TreeIterator<EObject> patternContents = rule
	// .eAllContents();
	// Boolean moved = false;
	// while (patternContents.hasNext()) {
	// EObject ruleObject = patternContents.next();
	// // target node instanceof GPObject
	// if (ruleObject instanceof GPObject && !moved) {
	// EClass targetClass2 = (EClass) ((GPObject) ruleObject)
	// .getType();
	// String targetClassName2 = "";
	// if (targetClass2.eIsProxy()) {
	// URI classURI = EcoreUtil
	// .getURI(targetClass2);
	// String fragment = classURI.fragment();
	// targetClassName2 = fragment.split("/")[fragment
	// .split("/").length - 1];
	// } else {
	// targetClassName2 = targetClass2.getName();
	// }
	// if (targetClassName2.equals(targetClass
	// .getName())) {
	// // ((GPObject)
	// // ruleObject).getAttributes().addAll(((GPObject)
	// // eObject).getAttributes());
	// EList<GTAttribute<EAttribute>> attributes =
	// (EList<GTAttribute<EAttribute>>) (EList<?>) (((GPObject) eObject)
	// .getAttributes());
	// for (int i = 0; i < attributes.size(); i++) {
	// GTAttribute<EAttribute> a = attributes
	// .get(i);
	// EAttribute exAttr = a.getExAttribute();
	// ((GPObject) ruleObject).getAttributes()
	// .add(a);
	// if (exAttr.eIsProxy()) {
	// URI exAttrURI = EcoreUtil
	// .getURI(exAttr);
	// String fragment = exAttrURI
	// .fragment();
	// String exAttrName = fragment
	// .split("/")[fragment
	// .split("/").length - 1];
	// EAttribute temp = helper
	// .getAttributeByName(
	// targetClass2,
	// exAttrName);
	// a.setExAttribute(temp);
	// } else {
	// a.setExAttribute(exAttr);
	// }
	// }
	// ((GPObject) ruleObject).getFields().addAll(
	// ((GPObject) eObject).getFields());
	// TreeIterator<EObject> patternContents2 = rule
	// .eAllContents();
	// // reroute GTLinks (and reset the
	// // ExReference) or remove them
	// while (patternContents2.hasNext()) {
	// EObject ruleObject2 = patternContents2
	// .next();
	// if (ruleObject2 instanceof GTLink
	// && ((GTLink) ruleObject2)
	// .getSourceNode()
	// .equals((GPObject) eObject)) {
	// EObject exRef = ((GTLink) ruleObject2)
	// .getExReference();
	// ((GTLink) ruleObject2)
	// .setSourceNode((GPObject) ruleObject);
	// if (exRef.eIsProxy()) {
	// URI exRefURI = EcoreUtil
	// .getURI(exRef);
	// String fragment = exRefURI
	// .fragment();
	// String exRefName = fragment
	// .split("/")[fragment
	// .split("/").length - 1];
	// EClass temp = helper
	// .getClass(targetClassName2);
	// for (EReference r : temp
	// .getEReferences()) {
	// if (r.getName().equals(
	// exRefName)) {
	// ((GTLink) ruleObject2)
	// .setExReference(r);
	// }
	// }
	// } else {
	// ((GTLink) ruleObject2)
	// .setExReference(exRef);
	// }
	// } else if (ruleObject2 instanceof GTLink
	// && ((GTLink) ruleObject2)
	// .getTargetNode()
	// .equals((GPObject) eObject)) {
	// objectsToRemove.add(ruleObject2);
	// }
	// }
	// moved = true;
	// }
	// }
	// // target node instanceof GTThisObject
	// else if (ruleObject instanceof GTThisObject
	// && !moved) {
	// EClass targetClass2 = (EClass) ((GTThisObject) ruleObject)
	// .getType();
	// String targetClassName2 = "";
	// if (targetClass2.eIsProxy()) {
	// URI classURI = EcoreUtil
	// .getURI(targetClass2);
	// String fragment = classURI.fragment();
	// targetClassName2 = fragment.split("/")[fragment
	// .split("/").length - 1];
	// } else {
	// targetClassName2 = targetClass2.getName();
	// }
	// if (targetClassName2.equals(targetClass
	// .getName())) {
	// // ((GTThisObject)
	// // ruleObject).getAttributes().addAll(((GPObject)
	// // eObject).getAttributes());
	// EList<GTAttribute<EAttribute>> attributes =
	// (EList<GTAttribute<EAttribute>>) (EList<?>) (((GPObject) eObject)
	// .getAttributes());
	// for (int i = 0; i < attributes.size(); i++) {
	// GTAttribute<EAttribute> a = attributes
	// .get(i);
	// EAttribute exAttr = a.getExAttribute();
	// ((GTThisObject) ruleObject)
	// .getAttributes().add(a);
	// if (exAttr.eIsProxy()) {
	// URI exAttrURI = EcoreUtil
	// .getURI(exAttr);
	// String fragment = exAttrURI
	// .fragment();
	// String exAttrName = fragment
	// .split("/")[fragment
	// .split("/").length - 1];
	// EAttribute temp = helper
	// .getAttributeByName(
	// targetClass2,
	// exAttrName);
	// a.setExAttribute(temp);
	// } else {
	// a.setExAttribute(exAttr);
	// }
	// }
	// ((GTThisObject) ruleObject).getFields()
	// .addAll(((GPObject) eObject)
	// .getFields());
	// TreeIterator<EObject> patternContents2 = rule
	// .eAllContents();
	// // reroute GTLinks (and reset the
	// // ExReference) or remove them
	// while (patternContents2.hasNext()) {
	// EObject ruleObject2 = patternContents2
	// .next();
	// if (ruleObject2 instanceof GTLink
	// && ((GTLink) ruleObject2)
	// .getSourceNode()
	// .equals((GPObject) eObject)) {
	// EObject exRef = ((GTLink) ruleObject2)
	// .getExReference();
	// ((GTLink) ruleObject2)
	// .setSourceNode((GTThisObject) ruleObject);
	// if (exRef.eIsProxy()) {
	// URI exRefURI = EcoreUtil
	// .getURI(exRef);
	// String fragment = exRefURI
	// .fragment();
	// String exRefName = fragment
	// .split("/")[fragment
	// .split("/").length - 1];
	// EClass temp = helper
	// .getClass(targetClassName2);
	// for (EReference r : temp
	// .getEReferences()) {
	// if (r.getName().equals(
	// exRefName)) {
	// ((GTLink) ruleObject2)
	// .setExReference(r);
	//
	// }
	// }
	// } else {
	// ((GTLink) ruleObject2)
	// .setExReference(exRef);
	// }
	// } else if (ruleObject2 instanceof GTLink
	// && ((GTLink) ruleObject2)
	// .getTargetNode()
	// .equals((GPObject) eObject)) {
	// objectsToRemove.add(ruleObject2);
	// }
	// }
	// moved = true;
	// }
	// }
	// }
	// // target class not present in the rule
	// if (!moved) {
	// TreeIterator<EObject> patternContents2 = rule
	// .eAllContents();
	// while (patternContents2.hasNext()) {
	// EObject ruleObject2 = patternContents2.next();
	// if (ruleObject2 instanceof GTLink
	// && (((GTLink) ruleObject2)
	// .getSourceNode().equals(
	// (GPObject) eObject) || ((GTLink) ruleObject2)
	// .getTargetNode().equals(
	// (GPObject) eObject))) {
	// objectsToRemove.add((GTLink) ruleObject2);
	// }
	// }
	// }
	// if (!objectsToRemove.isEmpty()) {
	// for (EObject o : objectsToRemove) {
	// EcoreUtil.remove(o);
	// }
	// }
	// try {
	// res.save(Collections.EMPTY_MAP);
	// } catch (IOException e) {
	// e.printStackTrace();
	// }
	// break;
	// }
	// // source node instanceof GTThisObject
	// else if (eObject instanceof GTThisObject) {
	// JOptionPane
	// .showMessageDialog(
	// null,
	// "Reason: the class that was removed is a This node in the rule\nPlease change the rule manually.",
	// "One or more rules could not be edited!",
	// JOptionPane.WARNING_MESSAGE);
	// }
	// }
	// }
	// }
	// }
}