package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Pattern;

import javax.swing.JOptionPane;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.ETypedElement;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.FeatureChange;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.presentation.EcoreEditor;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.gmf.runtime.diagram.core.services.ViewService;
import org.eclipse.gmf.runtime.notation.BasicCompartment;
import org.eclipse.gmf.runtime.notation.DecorationNode;
import org.eclipse.gmf.runtime.notation.Diagram;
import org.eclipse.gmf.runtime.notation.Node;
import org.eclipse.gmf.runtime.notation.Shape;
import org.eclipse.gmf.runtime.notation.View;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;

import de.ubt.ai1.modgraph.gt.gtEngine.GPElement;
import de.ubt.ai1.modgraph.gt.gtEngine.GPLink;
import de.ubt.ai1.modgraph.gt.gtEngine.GPObject;
import de.ubt.ai1.modgraph.gt.gtEngine.GPUnboundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GTAttribute;
import de.ubt.ai1.modgraph.gt.gtEngine.GTField;
import de.ubt.ai1.modgraph.gt.gtEngine.GTGraphPattern;
import de.ubt.ai1.modgraph.gt.gtEngine.GTNegativeApplicationCondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPath;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPostcondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPrecondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTStatus;
import de.ubt.ai1.modgraph.gt.gtEngine.GTThisObject;
import de.ubt.ai1.modgraph.gt.gtEngine.GraphTransformationRule;
import de.ubt.ai1.modgraph.gt.gtEngine.GtEngineFactory;
import de.ubt.ai1.modgraph.gt.gtEngine.NACBoundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.NACElement;
import de.ubt.ai1.modgraph.gt.gtEngine.NACLink;
import de.ubt.ai1.modgraph.gt.gtEngine.NACNode;
import de.ubt.ai1.modgraph.gt.gtEngine.NACObject;
import de.ubt.ai1.modgraph.gtEngine.diagram.part.GTDiagramEditorPlugin;

public class ExtractEClassConnector extends RefactoringConnector {
	private IEditorPart activeEditor;
	private String newClassName = "";
	private String newReferenceName = "";
	private String existingClassnewName = "";
	private EClass existingClass;
	private EClass extractedClass;
	private EReference newReference;

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		activeEditor = HandlerUtil.getActiveEditor(event);
		window = HandlerUtil.getActiveWorkbenchWindow(event);

		if (activeEditor instanceof EcoreEditor) {

			EcoreEditor ecoreEditor = (EcoreEditor) activeEditor;
			ecoreEditor.doSave(null);

			// Klasse aus welcher extrahiert wird und die zu extrahierenden
			// Elemente sind selektiert.

			TreeSelection sel = (TreeSelection) ecoreEditor.getSelection();

			List<EStructuralFeature> elementsToMove = new LinkedList<EStructuralFeature>();
			List<EOperation> operationsToMove = new LinkedList<EOperation>();
			Iterator<EObject> iter = sel.iterator();
			while (iter.hasNext()) {
				EObject o = iter.next();
				if (o instanceof EClass)
					existingClass = (EClass) o;
				else if (o instanceof EOperation) {
					operationsToMove.add((EOperation) o);
				} else if (o instanceof EReference || o instanceof EAttribute) {
					elementsToMove.add((EStructuralFeature) o);
				}
			}

			// ausgewaehltes Element pruefen
			if (existingClass != null) {
				EPackage pack = existingClass.getEPackage();
				Resource res = pack.eResource();

				newClassName = JOptionPane
						.showInputDialog("Enter the name for the extracted class.");
				newReferenceName = JOptionPane
						.showInputDialog("Enter the name for the reference to be created.");
				existingClassnewName = JOptionPane
						.showInputDialog("Optional new name for the existing class.");
				if (existingClassnewName == null
						|| existingClassnewName.equals("")) {
					existingClassnewName = existingClass.getName();
				}
				ChangeRecorder recorder = new ChangeRecorder(existingClass);

				// PERFORM REFACTORING:
				try {
					// refactoring Regel aufrufen
					extractedClass = refactoring.extractEClassNEW(
							existingClass, existingClassnewName, newClassName,
							newReferenceName,
							new BasicEList<EStructuralFeature>(elementsToMove),
							new BasicEList<EOperation>(operationsToMove));
				} catch (Throwable e) {
					// Fehlermeldung ausgeben, falls probleme auftreten sollten
					JOptionPane.showMessageDialog(null, "Refactoring failed.",
							"ERROR!", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
					return null;
				}
				// ressource speichern
				try {
					res.save(Collections.EMPTY_MAP);
				} catch (IOException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
				try {
					ChangeDescription description = recorder.endRecording();
					if (!(description.eContents().isEmpty())) {
						// PROPAGATION AUFRUFEN:
						propagateChanges(description, existingClass, res);

					}
				} catch (Throwable e) {
					e.printStackTrace();
				}
			}

		}
		return null;

	}

	@Override
	protected void applyToAffectedObjects(ENamedElement eNamedElement) {
		EList<Object> temp = new BasicEList<Object>();
		EList<EObject> objectsToMove = new BasicEList<EObject>();
		for (Resource gtRes : gtResources) {
			TreeIterator<EObject> it = gtRes.getAllContents();

			while (it.hasNext()) {
				EObject eObject = it.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					GraphTransformationRule<?, ?> rule = (GraphTransformationRule<?, ?>) eObject;
					objectsToMove = propagation.propagateExtractClassNEW(extractedClass,
							existingClass, rule);
					try {
						gtRes.save(Collections.EMPTY_MAP);
					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					break;
				}
			}
		}
		// ALT:
		for (Resource res : gtResources) {
			temp.add((Object) res);
			EcoreUtil.resolveAll(res);
			TreeIterator<EObject> iter = res.getAllContents();
//			EList<EObject> objectsToMove = new BasicEList<EObject>();
			EList<EObject> objectsToRemove = new BasicEList<EObject>();
			GraphTransformationRule rule = null;
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				 if (eObject instanceof GraphTransformationRule<?, ?>) {
				 rule = (GraphTransformationRule) eObject;
				 
				// EClass targetClass = null;
				// for (EClass c : refactoring.getReferenceToEClass())
				// if (c.getName().equals(newClassName))
				// targetClass = c;
				// objectsToMove = propagation.propagateExtractClass(
				// targetClass, existingClass,
				// (GraphTransformationRule) eObject);
				//
				 }
				// else
				if (eObject instanceof GTField
						&& ((GTField) eObject).getKind().getName()
								.equals("OCL_CONSTRAINT")) {
					String containerName = "OCL_CONSTRAINT";
					String expression = ((GTField) eObject).getContent();
					String newExpression = propagation
							.updateOCLExpressionWithUserInteraction(
									rule.getName(), expression,
									existingClass.getName(), containerName);
					if (newExpression != null)
						((GTField) eObject).setContent(newExpression);
					if (newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				} else if (eObject instanceof GTPath) {
					String containerName = "GTPath";
					String expression = ((GTPath) eObject).getExpression();
					String newExpression = propagation
							.updateOCLExpressionWithUserInteraction(
									rule.getName(), expression,
									existingClass.getName(), containerName);
					if (newExpression != null)
						((GTPath) eObject).setExpression(newExpression);
					if (newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				} else if (eObject instanceof GTPrecondition) {
					String containerName = "GTPrecondition";
					String expression = ((GTPrecondition) eObject)
							.getCondition();
					String newExpression = propagation
							.updateOCLExpressionWithUserInteraction(
									rule.getName(), expression,
									existingClass.getName(), containerName);
					if (newExpression != null)
						((GTPrecondition) eObject).setCondition(newExpression);
					if (newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				} else if (eObject instanceof GTPostcondition) {
					String containerName = "GTPostcondition";
					String expression = ((GTPostcondition) eObject)
							.getCondition();
					String newExpression = propagation
							.updateOCLExpressionWithUserInteraction(
									rule.getName(), expression,
									existingClass.getName(), containerName);
					if (newExpression != null)
						((GTPostcondition) eObject).setCondition(newExpression);
					if (newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				} else if (eObject instanceof Shape
						&& objectsToRemove.contains(((Shape) eObject)
								.getElement()))
					objectsToRemove.add(eObject);
				else if (eObject instanceof Diagram && !objectsToMove.isEmpty()
						) {
					EObject patternNodeContainer = null;
					EList<EObject> nacNodeContainers = new BasicEList<EObject>();
					TreeIterator<EObject> iter2 = res.getAllContents();
					while (iter2.hasNext()) {
						EObject eObject2 = iter2.next();
						if (eObject2 instanceof DecorationNode
								&& ((Node) eObject2).getType().equals("5001")) {
							patternNodeContainer = eObject2;
						}
						if (eObject2 instanceof DecorationNode
								&& ((Node) eObject2).getType().equals("5017")) {
							nacNodeContainers.add(eObject2);
						}
					}
					EObject nodeElement = null;
					if (rule != null) {
						for (EObject o : rule.getPattern().getElements()) {
							if (o instanceof GTNode
									&& ((GTNode) o).getName().equals(
											"newNode")) {
								nodeElement = o;
								Node patternNode = ViewService
										.createNode(
												(View) patternNodeContainer,
												nodeElement,
												"2011",
												GTDiagramEditorPlugin.DIAGRAM_PREFERENCES_HINT);
								for (EObject object : objectsToMove) {
									if (((Node) object.eContainer()
											.eContainer().eContainer())
											.getType().equals("5001")) {
										for (Object o2 : patternNode
												.getPersistedChildren()) {
											if (o2 instanceof BasicCompartment
													&& (((Shape) object)
															.getElement() instanceof GTAttribute && !((GTAttribute) ((Shape) object)
															.getElement())
															.getOperator()
															.getLiteral()
															.equals("="))) {
												((BasicCompartment) o2)
														.getPersistedChildren()
														.add(object);
												break;
											} else if (o2 instanceof BasicCompartment
													&& ((((Shape) object)
															.getElement() instanceof GTAttribute && ((GTAttribute) ((Shape) object)
															.getElement())
															.getOperator()
															.getLiteral()
															.equals("=")) || (((Shape) object)
															.getElement() instanceof GTField && ((GTField) ((Shape) object)
															.getElement())
															.getKind()
															.getName()
															.equals("OPERATION_CALL")))) {
												((BasicCompartment) o2)
														.getPersistedChildren()
														.add(object);
											}
										}
									}
								}
							}
						}
						int i = 0;
						for (Object tempNAC : rule.getNegativePattern()) {
							i++;
							GTNegativeApplicationCondition nac = (GTNegativeApplicationCondition) tempNAC;
							for (EObject o : nac.getElements()) {
								if (o instanceof GTNode
										&& ((GTNode) o).getName().equals(
												"newNode")) {
									nodeElement = o;
									int j = 0;
									for (EObject nacNodeContainer : nacNodeContainers) {
										j++;
										if (i == j) {
											Node nacNode = ViewService
													.createNode(
															(View) nacNodeContainer,
															nodeElement,
															"2017",
															GTDiagramEditorPlugin.DIAGRAM_PREFERENCES_HINT);
											for (EObject object : objectsToMove) {
												if (((Node) object.eContainer()
														.eContainer()
														.eContainer())
														.getType().equals(
																"5017")) {
													for (Object o2 : nacNode
															.getPersistedChildren()) {
														if (o2 instanceof BasicCompartment
																&& (((Shape) object)
																		.getElement() instanceof GTAttribute && !((GTAttribute) ((Shape) object)
																		.getElement())
																		.getOperator()
																		.getLiteral()
																		.equals("="))
																&& (((Shape) object)
																		.getElement()
																		.eContainer()
																		.eContainer()
																		.equals(nac))) {
															((BasicCompartment) o2)
																	.getPersistedChildren()
																	.add(object);
															break;
														} else if (o2 instanceof BasicCompartment
																&& ((((Shape) object)
																		.getElement() instanceof GTAttribute && ((GTAttribute) ((Shape) object)
																		.getElement())
																		.getOperator()
																		.getLiteral()
																		.equals("=")) || (((Shape) object)
																		.getElement() instanceof GTField && ((GTField) ((Shape) object)
																		.getElement())
																		.getKind()
																		.getName()
																		.equals("OPERATION_CALL")))
																&& (((Shape) object)
																		.getElement()
																		.eContainer()
																		.eContainer()
																		.equals(nac))) {
															((BasicCompartment) o2)
																	.getPersistedChildren()
																	.add(object);
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
			propagation.cleanUpAndSave(objectsToRemove, res);

			try {
				res.save(Collections.EMPTY_MAP);
			} catch (IOException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
			URI resURI = res.getURI();
			String fragment = resURI.toString();
			String[] resRuleNameArray = fragment.split("/");
			String resRuleName = resRuleNameArray[resRuleNameArray.length - 1];
			int response = JOptionPane.showConfirmDialog(null,
					"Would you like to validate the GT rule " + resRuleName
							+ "?", "Attention", JOptionPane.YES_NO_OPTION);
			if (response == 0) {
				try {
					boolean result = callValidation(res);
					if (result == false) {
						JOptionPane
								.showMessageDialog(
										null,
										"Errors exist in the rule "
												+ resRuleName
												+ "\nPlease check it manually.",
										"Validation failed!",
										JOptionPane.ERROR_MESSAGE);
					}
				} catch (Throwable e) {
					e.printStackTrace();
				}
			}
		}
	}

	// protected boolean propagateChanges2(ChangeDescription description,
	// ENamedElement eNamedElement) {
	// EList<FeatureChange> changes = description.getObjectChanges().get(
	// eNamedElement);
	// EList<EObject> addedObjects = description.getObjectsToAttach();
	// EList<EObject> removedObjects = description.getObjectsToDetach();
	// getGTResources(eNamedElement);
	// if (changes != null)
	// for (FeatureChange change : changes) {
	// System.out.println(change + " Name: " + change.getFeatureName()
	// + " Value: " + change.getDataValue());
	// applyToAffectedObjects2(eNamedElement);
	// }
	// if (addedObjects != null)
	// for (EObject addedObj : addedObjects) {
	// System.out.println(addedObj);
	//
	// }
	// if (removedObjects != null)
	// for (EObject removedObj : removedObjects) {
	// System.out.println(removedObj);
	// }
	// return false;
	// }

	@SuppressWarnings({ "rawtypes", "unchecked" })
	protected void applyToAffectedObjects2(ENamedElement eNamedElement) {
		// ValidationMenuHandler vmh = new ValidationMenuHandler();
		for (Resource res : gtResources) {
			TreeIterator<EObject> iter = res.getAllContents();
			EList<EObject> objectsToRemove = new BasicEList<EObject>();
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					EClass targetClass = null;
					for (EClass c : refactoring.getReferenceToEClass())
						if (c.getName().equals(newClassName))
							targetClass = c;
					GTGraphPattern pattern = ((GraphTransformationRule) eObject)
							.getPattern();
					GTNode sourceNode = null;
					for (GPElement gpu : pattern.getElements()) {
						if (gpu instanceof GPUnboundNode
								&& ((EClass) ((GPUnboundNode) gpu).getType())
										.getName().equals(
												eNamedElement.getName()))
							sourceNode = (GPUnboundNode) gpu;
						else if (gpu instanceof GTThisObject
								&& ((EClass) ((GTThisObject) gpu).getType())
										.getName().equals(
												eNamedElement.getName()))
							sourceNode = (GTThisObject) gpu;
					}
					if (sourceNode != null) {
						GPObject newGPObject = GtEngineFactory.eINSTANCE
								.createGPObject();
						newGPObject.setName(newClassName);
						newGPObject.setType(targetClass);
						newGPObject.setStatus(GTStatus.PRESERVE);
						pattern.getElements().add(newGPObject);

						// IElementType type = XXXElementTypes.Xxxx_1001; // the
						// primitive objects IElementType you want to create
						// CreateViewRequest createViewRequest =
						// CreateViewRequestFactory.getCreateShapeRequest (type,
						// XXXDiagramEditorPlugin.DIAGRAM_PREFERENCES_HINT);
						// org.eclipse.gef.commands.Command createCommand =
						// xxxEditPart.getCommand (createViewRequest);

						// CreateUnspecifiedTypeRequest request = new
						// CreateUnspecifiedTypeRequest(
						// Collections.singletonList(EcoreElementTypes.EClass_1001),
						// diagramEditPart.getDiagramPreferencesHint());
						//
						// Command command =
						// diagramEditPart.getCommand(request);
						// command.execute();
						// Object newObject = request.getNewObject(); // these
						// are the newly created objects

						// Node node = ViewFactory.

						List<EObject> tempGTA = new ArrayList<>();
						for (EAttribute ea : targetClass.getEAttributes())
							for (GTAttribute gta : sourceNode.getAttributes())
								if (gta.getExAttribute().eIsProxy()
										&& propagation.getNameFromURI(
												gta.getExAttribute()).equals(
												ea.getName())) {
									tempGTA.add(gta);
									gta.setExAttribute(ea);
								}
						newGPObject.getAttributes().addAll(
								(Collection<? extends GTAttribute<?>>) tempGTA);
						markShapesToRemove(res, tempGTA, objectsToRemove);
						List<EObject> tempGTF = new ArrayList<>();
						for (EOperation eo : targetClass.getEOperations())
							for (GTField gtf : sourceNode.getFields())
								if (gtf.getKind().getName()
										.equals("OPERATION_CALL")) {
									String content = ((GTField) gtf)
											.getContent();
									String operationName = content
											.split(Pattern.quote("("))[0];
									if (operationName.equals(eo.getName()))
										tempGTF.add(gtf);
								}
						newGPObject.getFields().addAll(
								(Collection<? extends GTField>) tempGTF);
						markShapesToRemove(res, tempGTF, objectsToRemove);
						for (EReference er : ((EClass) eNamedElement)
								.getEReferences())
							if (er.getName().equals(newClassName)) {
								GPLink newGPLink = GtEngineFactory.eINSTANCE
										.createGPLink();
								newGPLink.setExReference(er);
								newGPLink.setSourceNode(sourceNode);
								newGPLink.setTargetNode(newGPObject);
								newGPLink.setStatus(GTStatus.PRESERVE);
								sourceNode.getOutgoingEdges().add(newGPLink);
								newGPObject.getIncomingEdges().add(newGPLink);
								pattern.getElements().add(newGPLink);
							}
					}
					List<GTNegativeApplicationCondition> nacList = ((GraphTransformationRule) eObject)
							.getNegativePattern();
					for (GTNegativeApplicationCondition nac : nacList) {
						NACNode sourceNodeNAC = null;
						for (NACElement nacNode : nac.getElements()) {
							if (nacNode instanceof NACBoundNode
									&& ((EClass) ((NACBoundNode) nacNode)
											.getGPUnboundNode().getType())
											.getName().equals(
													eNamedElement.getName()))
								sourceNodeNAC = (NACNode) nacNode;
							else if (nacNode instanceof NACObject
									&& ((EClass) ((NACObject) nacNode)
											.getType()).getName().equals(
											eNamedElement.getName()))
								sourceNodeNAC = (NACNode) nacNode;
							else if (nacNode instanceof GTThisObject
									&& ((EClass) ((GTThisObject) nacNode)
											.getType()).getName().equals(
											eNamedElement.getName()))
								sourceNodeNAC = (NACNode) nacNode;
						}
						if (sourceNodeNAC != null) {
							NACObject newNACObject = GtEngineFactory.eINSTANCE
									.createNACObject();
							newNACObject.setName(newClassName);
							newNACObject.setType(targetClass);
							nac.getElements().add(newNACObject);
							List<EObject> tempGTA = new ArrayList<>();
							for (EAttribute ea : targetClass.getEAttributes())
								for (GTAttribute gta : sourceNodeNAC
										.getAttributes())
									if (gta.getExAttribute().eIsProxy()
											&& propagation.getNameFromURI(
													gta.getExAttribute())
													.equals(ea.getName())) {
										tempGTA.add(gta);
										gta.setExAttribute(ea);
									}
							newNACObject
									.getAttributes()
									.addAll((Collection<? extends GTAttribute<?>>) tempGTA);
							markShapesToRemove(res, tempGTA, objectsToRemove);
							for (EReference er : ((EClass) eNamedElement)
									.getEReferences())
								if (er.getName().equals(newClassName)) {
									NACLink newNACLink = GtEngineFactory.eINSTANCE
											.createNACLink();
									newNACLink.setExReference(er);
									newNACLink.setSourceNode(sourceNodeNAC);
									newNACLink.setTargetNode(newNACObject);
									sourceNodeNAC.getOutgoingEdges().add(
											newNACLink);
									newNACObject.getIncomingEdges().add(
											newNACLink);
									nac.getElements().add(newNACLink);
								}
						}
					}
				}
				break;
			}
			propagation.cleanUpAndSave(objectsToRemove, res);
			// vmh.initValidation(gtFiles.get(gtResources.indexOf(res)), window,
			// new NullProgressMonitor());
		}
	}

	private void markShapesToRemove(Resource res, List<EObject> elements,
			List<EObject> objectsToRemove) {
		TreeIterator<EObject> iter2 = res.getAllContents();
		while (iter2.hasNext()) {
			EObject eObject = iter2.next();
			if (eObject instanceof Shape
					&& elements.contains(((Shape) eObject).getElement()))
				objectsToRemove.add((Shape) eObject);
		}
	}
}