package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.io.IOException;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JOptionPane;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.presentation.EcoreEditor;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.gmf.runtime.diagram.core.services.ViewService;
import org.eclipse.gmf.runtime.notation.BasicCompartment;
import org.eclipse.gmf.runtime.notation.DecorationNode;
import org.eclipse.gmf.runtime.notation.Diagram;
import org.eclipse.gmf.runtime.notation.Node;
import org.eclipse.gmf.runtime.notation.Shape;
import org.eclipse.gmf.runtime.notation.View;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;

import de.ubt.ai1.modgraph.gt.gtEngine.GTAttribute;
import de.ubt.ai1.modgraph.gt.gtEngine.GTField;
import de.ubt.ai1.modgraph.gt.gtEngine.GTNegativeApplicationCondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPath;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPostcondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPrecondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GraphTransformationRule;
import de.ubt.ai1.modgraph.gtEngine.diagram.part.GTDiagramEditorPlugin;

public class ExtractSubClassConnector extends RefactoringConnector {
	private IEditorPart activeEditor;
	private String newName;
	private EClass sourceClass;
	
	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		activeEditor = HandlerUtil.getActiveEditor(event);
		window = HandlerUtil.getActiveWorkbenchWindow(event);

		if (activeEditor instanceof EcoreEditor) {
			EcoreEditor ecoreEditor = (EcoreEditor) activeEditor;
			ecoreEditor.doSave(null);
			TreeSelection sel = (TreeSelection) ecoreEditor.getSelection();
			Object obj = sel.getFirstElement();

			if (obj instanceof EClass) {
				sourceClass = (EClass) obj;
				EPackage pack = sourceClass.getEPackage();
				Resource res = pack.eResource();
				List<String> classNames = new LinkedList<String>();
				List<EClassifier> classes = pack.getEClassifiers();
				for (EClassifier c : classes) {
					if (c instanceof EClass)
						classNames.add(c.getName());
				}
				newName = JOptionPane
						.showInputDialog("Enter the name for the extracted subclass.");
				if (classNames.contains(newName)) {
					JOptionPane.showMessageDialog(null,
							"A class with that name already exists.", "ERROR!",
							JOptionPane.ERROR_MESSAGE);
					return null;
				}
				if(newName.equals("")){
					JOptionPane.showMessageDialog(null,
							"Selected name is not valid.", "ERROR!",
							JOptionPane.ERROR_MESSAGE);
					return null;					
				}

				initRefactoring(sourceClass.getEPackage().eContents());

				int cond = 0;
				EList<String> attributes = new BasicEList<String>();
				List<String> attributeNames = new LinkedList<String>();
				for (EAttribute a : sourceClass.getEAttributes())
					attributeNames.add(a.getName());
				if (!attributeNames.isEmpty()) {
					cond = JOptionPane.showConfirmDialog(null,
							"Add an attribute to be moved?", "",
							JOptionPane.YES_NO_OPTION);
					while (cond == JOptionPane.YES_OPTION
							&& !attributeNames.isEmpty()) {
						String temp = getSelectedFromDialog(attributeNames,
								"Choose an attribute.");
						attributes.add(helper.getAttributeByName(sourceClass,
								temp).getName());
						attributeNames.remove(temp);
						cond = JOptionPane.NO_OPTION;
						if (!attributeNames.isEmpty()) {
							cond = JOptionPane.showConfirmDialog(null,
									"Add an attribute to be moved?", "",
									JOptionPane.YES_NO_OPTION);
						}
					}
				}

				EList<String> operations = new BasicEList<String>();
				List<String> operationNames = new LinkedList<String>();
				for (EOperation a : sourceClass.getEOperations())
					operationNames.add(a.getName());
				if (!operationNames.isEmpty()) {
					cond = JOptionPane.showConfirmDialog(null,
							"Add an operation to be moved?", "",
							JOptionPane.YES_NO_OPTION);
					while (cond == JOptionPane.YES_OPTION
							&& !operationNames.isEmpty()) {
						String temp = getSelectedFromDialog(operationNames,
								"Choose an operation.");
						operations.add(helper.getOperationByName(sourceClass,
								temp).getName());
						operationNames.remove(temp);
						cond = JOptionPane.NO_OPTION;
						if (!operationNames.isEmpty()) {
							cond = JOptionPane.showConfirmDialog(null,
									"Add an operation to be moved?", "",
									JOptionPane.YES_NO_OPTION);
						}
					}
				}
				ChangeRecorder recorder = new ChangeRecorder(sourceClass);
				try {
					refactoring.extractSubClass(sourceClass.getName(), newName,
							operations, attributes);
				} catch (Throwable e) {
					JOptionPane.showMessageDialog(null, "Refactoring failed.",
							"ERROR!", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
					return null;
				}
				try {
					res.save(Collections.EMPTY_MAP);
				} catch (IOException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
				try {
					ChangeDescription description = recorder.endRecording();
					if (!(description.eContents().isEmpty())) {
						propagateChanges(description, sourceClass, res);
					}					
				} catch (Throwable e){
					e.printStackTrace();
				}
			}

		}
		return null;

	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	@Override
	protected void applyToAffectedObjects(ENamedElement eNamedElement) {
		EList<Object> temp = new BasicEList<Object>();
		for (Resource res : gtResources) {
			temp.add((Object) res);		
			EcoreUtil.resolveAll(res);			
			TreeIterator<EObject> iter = res.getAllContents();
			EList<EObject> objectsToMove = new BasicEList<EObject>();
			EList<EObject> objectsToRemove = new BasicEList<EObject>();
			GraphTransformationRule rule = null;
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					EClass targetClass = null;
					rule = ((GraphTransformationRule) eObject);
					for(EClass c : refactoring.getReferenceToEClass())
						if(c.getName().equals(newName))
							targetClass = c;
					objectsToMove = propagation.propagateExtractSubClass(targetClass, sourceClass, (GraphTransformationRule) eObject);
				}
				else if (eObject instanceof GTField	&& ((GTField) eObject).getKind().getName().equals("OCL_CONSTRAINT")) {
					String containerName = "OCL_CONSTRAINT";
					String expression = ((GTField) eObject).getContent();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(rule.getName(), expression, sourceClass.getName(), containerName);
					if(newExpression != null)
						((GTField) eObject).setContent(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if (eObject instanceof GTPath) {
					String containerName = "GTPath";
					String expression = ((GTPath) eObject).getExpression();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(rule.getName(), expression, sourceClass.getName(), containerName);
					if(newExpression != null)
						((GTPath) eObject).setExpression(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPrecondition){
					String containerName = "GTPrecondition";
					String expression = ((GTPrecondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(rule.getName(), expression, sourceClass.getName(), containerName);
					if(newExpression != null)
						((GTPrecondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPostcondition){
					String containerName = "GTPostcondition";
					String expression = ((GTPostcondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(rule.getName(), expression, sourceClass.getName(), containerName);
					if(newExpression != null)
						((GTPostcondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof Shape && objectsToRemove.contains(((Shape)eObject).getElement()))
					objectsToRemove.add(eObject);
				else if (eObject instanceof Diagram && !objectsToMove.isEmpty()) {
					EObject patternNodeContainer = null;
					EList<EObject> nacNodeContainers = new BasicEList<EObject>();
					TreeIterator<EObject> iter2 = res.getAllContents();
					while (iter2.hasNext()) {
						EObject eObject2 = iter2.next();
						if(eObject2 instanceof DecorationNode && ((Node)eObject2).getType().equals("5001")){
							patternNodeContainer = eObject2;
						}
						if(eObject2 instanceof DecorationNode && ((Node)eObject2).getType().equals("5017")){
							nacNodeContainers.add(eObject2);
						}
					}					
					EObject nodeElement = null;
					if(rule != null){
						for(EObject o : rule.getPattern().getElements()){
							if(o instanceof GTNode && ((GTNode)o).getName().equals(newName)){
								nodeElement = o;
								Node patternNode = ViewService.createNode((View) patternNodeContainer, nodeElement, "2011", GTDiagramEditorPlugin.DIAGRAM_PREFERENCES_HINT);
								for(EObject object : objectsToMove){
									if(((Node) object.eContainer().eContainer().eContainer()).getType().equals("5001")){
										for(Object o2 : patternNode.getPersistedChildren()){
											if(o2 instanceof BasicCompartment 
												&& (((Shape)object).getElement() instanceof GTAttribute && !((GTAttribute) ((Shape)object).getElement()).getOperator().getLiteral().equals("=")))
											{
												((BasicCompartment)o2).getPersistedChildren().add(object);
												break;
											}
											else if(o2 instanceof BasicCompartment 
												&& ((((Shape)object).getElement() instanceof GTAttribute && ((GTAttribute) ((Shape)object).getElement()).getOperator().getLiteral().equals("="))
												|| (((Shape)object).getElement() instanceof GTField && ((GTField) ((Shape)object).getElement()).getKind().getName().equals("OPERATION_CALL"))))
											{
												((BasicCompartment)o2).getPersistedChildren().add(object);
											}
										}
									}
								}
							}
						}
						int i = 0;
						for(Object tempNAC : rule.getNegativePattern()){
							i++;
							GTNegativeApplicationCondition nac = (GTNegativeApplicationCondition) tempNAC;
							for(EObject o : nac.getElements()){
								if(o instanceof GTNode && ((GTNode)o).getName().equals(newName)){
									nodeElement = o;
									int j = 0;
									for(EObject nacNodeContainer : nacNodeContainers){
										j++;
										if(i==j){
											Node nacNode = ViewService.createNode((View) nacNodeContainer, nodeElement, "2017", GTDiagramEditorPlugin.DIAGRAM_PREFERENCES_HINT);
											for(EObject object : objectsToMove){
												if(((Node) object.eContainer().eContainer().eContainer()).getType().equals("5017")){
													for(Object o2 : nacNode.getPersistedChildren()){
														if(o2 instanceof BasicCompartment 
															&& (((Shape)object).getElement() instanceof GTAttribute && !((GTAttribute) ((Shape)object).getElement()).getOperator().getLiteral().equals("="))
															&& (((Shape)object).getElement().eContainer().eContainer().equals(nac)))
														{
															((BasicCompartment)o2).getPersistedChildren().add(object);
															break;
														}
														else if(o2 instanceof BasicCompartment 
															&& ((((Shape)object).getElement() instanceof GTAttribute && ((GTAttribute) ((Shape)object).getElement()).getOperator().getLiteral().equals("="))
															|| (((Shape)object).getElement() instanceof GTField && ((GTField) ((Shape)object).getElement()).getKind().getName().equals("OPERATION_CALL")))
															&& (((Shape)object).getElement().eContainer().eContainer().equals(nac)))
														{
															((BasicCompartment)o2).getPersistedChildren().add(object);
														}
													}
												}
											}											
										}
									}
								}
							}
						}
					}					
				}
			}
			propagation.cleanUpAndSave(objectsToRemove, res);
			
			URI resURI = res.getURI();
			String fragment = resURI.toString();
			String[] resRuleNameArray = fragment.split("/");
			String resRuleName = resRuleNameArray[resRuleNameArray.length-1];			
			int response = JOptionPane.showConfirmDialog(null,
					"Would you like to validate the GT rule " + resRuleName + "?", "Attention",
					JOptionPane.YES_NO_OPTION);
			if(response == 0){
				try{
					boolean result = callValidation(res);
					if(result == false){
						JOptionPane.showMessageDialog(null,
								"Errors exist in the rule " + resRuleName + "\nPlease check it manually.", "Validation failed!",
								JOptionPane.ERROR_MESSAGE);
					}
				} catch (Throwable e){
					e.printStackTrace();
				}
			}
		}
	}
}