package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.channels.FileChannel;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JOptionPane;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IHandler;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.FeatureChange;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.presentation.EcoreEditor;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.gmf.runtime.notation.Edge;
import org.eclipse.gmf.runtime.notation.Shape;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;

import de.ubt.ai1.modGraph.example.eRefactoring.ERefactoringFactory;
import de.ubt.ai1.modgraph.gt.gtEngine.GPBoundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GPElement;
import de.ubt.ai1.modgraph.gt.gtEngine.GPUnboundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GTField;
import de.ubt.ai1.modgraph.gt.gtEngine.GTLink;
import de.ubt.ai1.modgraph.gt.gtEngine.GTNegativeApplicationCondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPath;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPostcondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPrecondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTThisObject;
import de.ubt.ai1.modgraph.gt.gtEngine.GraphTransformationRule;
import de.ubt.ai1.modgraph.gt.gtEngine.NACBoundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.NACElement;
import de.ubt.ai1.modgraph.gt.gtEngine.NACObject;
import de.ubt.ai1.modgraph.gt.gtEngine.NACParameter;

public class HideDelegateConnector extends RefactoringConnector implements
		IHandler {
	private IEditorPart activeEditor;
	private String serverName;
	private String delegateName;
	private String clientName;
	private EClass sourceClass;
	private EClass targetClass;
	private String refName = "";
	private int i = 0;
	private Resource res;

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		activeEditor = HandlerUtil.getActiveEditor(event);
		window = HandlerUtil.getActiveWorkbenchWindow(event);
		
		if (activeEditor instanceof EcoreEditor) {
			EcoreEditor ecoreEditor = (EcoreEditor) activeEditor;
			ecoreEditor.doSave(null);
			TreeSelection sel = (TreeSelection) ecoreEditor.getSelection();
			Object obj = sel.getFirstElement();

			if (obj instanceof EClass) {
				EClass delegate = (EClass) obj;
				res= delegate.eResource();
				EPackage pack = delegate.getEPackage();
				List<String> classNames = new LinkedList<String>();
				List<EClassifier> classes = pack.getEClassifiers();
				for (EClassifier c : classes) {
					if (c instanceof EClass)
						classNames.add(c.getName());
				}

				delegateName = delegate.getName();
				classNames.remove(delegateName);
				serverName = getSelectedFromDialog(classNames,
						"Choose the server class (Middle Man).");
				classNames.remove(serverName);

				clientName = getSelectedFromDialog(classNames,
						"Choose the client class.");
				classNames.remove(clientName);

				initRefactoring(pack.eContents());
				targetClass = delegate;
				sourceClass = helper.getClass(clientName);
				for(EReference r : sourceClass.getEReferences())
					if(r.getEType().getName().equals(targetClass.getName()))
						refName = r.getName();
				ChangeRecorder recorder = new ChangeRecorder(sourceClass);
				try {
					refactoring.hideDelegate(serverName, delegate.getName(),
							clientName);
				} catch (Throwable e) {
					JOptionPane.showMessageDialog(null, "Refactoring failed.",
							"ERROR!", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
					return null;
				}
				try {
					pack.eResource().save(Collections.EMPTY_MAP);
				} catch (IOException e) {
					e.printStackTrace();
				}
				ChangeDescription description = recorder.endRecording();
				if (!(description.eContents().isEmpty())) {
					propagateChanges(description, sourceClass,res);
				}				
			}
		}
		return null;
	}

	/**
	 * If the server and delegate classes are in the rule and server has a link
	 * to delegate removes the link.
	 */
	@SuppressWarnings({ "rawtypes" })
	@Override
	protected void applyToAffectedObjects(ENamedElement eNamedElement) {
		for (Resource res : gtResources) {
			TreeIterator<EObject> iter = res.getAllContents();
			EList<EObject> objectsToRemove = new BasicEList<EObject>();
			String ruleName = "";
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					ruleName = ((GraphTransformationRule) eObject).getName();
					try {
						propagation.propagateHideDeleg_ReplDelegWInheritInit(sourceClass, targetClass, (GraphTransformationRule) eObject);
					} catch (Throwable e1) {
						e1.printStackTrace();
						break;
					}
				} 
				else if (eObject instanceof GTField	&& ((GTField) eObject).getKind().getName().equals("OCL_CONSTRAINT")) {
					String containerName = "OCL_CONSTRAINT";
					String expression = ((GTField) eObject).getContent();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, refName, containerName);
					if(newExpression != null)
						((GTField) eObject).setContent(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if (eObject instanceof GTPath) {
					String containerName = "GTPath";
					String expression = ((GTPath) eObject).getExpression();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, refName, containerName);
					if(newExpression != null)
						((GTPath) eObject).setExpression(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPrecondition){
					String containerName = "GTPrecondition";
					String expression = ((GTPrecondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, refName, containerName);
					if(newExpression != null)
						((GTPrecondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPostcondition){
					String containerName = "GTPostcondition";
					String expression = ((GTPostcondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, refName, containerName);
					if(newExpression != null)
						((GTPostcondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof Shape && objectsToRemove.contains(((Shape)eObject).getElement()))
					objectsToRemove.add(eObject);
			}
			propagation.cleanUpAndSave(objectsToRemove, res);
			
			URI resURI = res.getURI();
			String fragment = resURI.toString();
			String[] resRuleNameArray = fragment.split("/");
			String resRuleName = resRuleNameArray[resRuleNameArray.length-1];			
			int response = JOptionPane.showConfirmDialog(null,
					"Would you like to validate the GT rule " + resRuleName + "?", "Attention",
					JOptionPane.YES_NO_OPTION);
			if(response == 0){
				try{
					boolean result = callValidation(res);
					if(result == false){
						JOptionPane.showMessageDialog(null,
								"Errors exist in the rule " + resRuleName + "\nPlease check it manually.", "Validation failed!",
								JOptionPane.ERROR_MESSAGE);
					}
				} catch (Throwable e){
					e.printStackTrace();
				}
			}
		}
	}
	
	
	// for runtime tests
	
	@SuppressWarnings("unused")
	private void test(EClass delegate) throws ExecutionException{
		execute2(delegate);
	}
	
	private Object execute2(EClass delegate) throws ExecutionException {
		EPackage pack = delegate.getEPackage();
		delegateName = delegate.getName();
		serverName = "Group";
		clientName = "BugTracker";

		refactoring = ERefactoringFactory.eINSTANCE.createRefactoring();
		helper = ERefactoringFactory.eINSTANCE.createHelper();
		propagation = ERefactoringFactory.eINSTANCE.createPropagation();
		initRefactoring(pack.eContents());
		
		targetClass = delegate;
		sourceClass = helper.getClass(clientName);
		for(EReference r : sourceClass.getEReferences())
			if(r.getEType().getName().equals(targetClass.getName()))
				refName = r.getName();
		ChangeRecorder recorder = new ChangeRecorder(sourceClass);
		long preRefactoringNano = System.nanoTime();
		try {
			refactoring.hideDelegate(serverName, delegate.getName(),
					clientName);
		} catch (Throwable e) {
			JOptionPane.showMessageDialog(null, "Refactoring failed.",
					"ERROR!", JOptionPane.ERROR_MESSAGE);
			e.printStackTrace();
			return null;
		}
//		for(EStructuralFeature es : sourceClass.getEStructuralFeatures()){
//			if(es instanceof EReference && es.getEType().getName().equals(delegateName)){
//				es.setEType(null);
//				sourceClass.getEStructuralFeatures().remove(es);
//				break;
//			}
//		}
//		try {
//			pack.eResource().save(Collections.EMPTY_MAP);
//		} catch (IOException e) {
//			e.printStackTrace();
//		}		
		try {
			pack.eResource().save(Collections.EMPTY_MAP);
		} catch (IOException e) {
			e.printStackTrace();
		}
		long prePropagationNano = System.nanoTime();
		ChangeDescription description = recorder.endRecording();
		if (!(description.eContents().isEmpty())) {
			propagateChanges(description, sourceClass,res);
//			propagateChanges2(description, sourceClass);
		}
		long finalNano = System.nanoTime();
		
		try(PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter("C:/RemoveMidMan_Refactoring.txt", true)))) {
			out.println((prePropagationNano - preRefactoringNano));
		}catch (IOException e) {}			
		try(PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter("C:/RemoveMidMan_Propagation.txt", true)))) {
			out.println((finalNano - prePropagationNano));
		}catch (IOException e) {}
		
		// For runtime testing only
		File source = new File("C:\\Users\\Duemmel\\Desktop\\Projekt\\TimeTestModels\\BTHideDelegate\\BugModel.ecore");
		File dest = new File("C:\\Users\\Duemmel\\runtime-EclipseApplication2\\Bugtracker\\model\\BugModel.ecore");
		FileChannel inputChannel = null;
		FileChannel outputChannel = null;
		try {
			inputChannel = new FileInputStream(source).getChannel();
			outputChannel = new FileOutputStream(dest).getChannel();
			outputChannel.transferFrom(inputChannel, 0, inputChannel.size());
		} catch (Throwable e) {	e.printStackTrace();} 
		finally {
			try {
				inputChannel.close();
				outputChannel.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		source = new File("C:\\Users\\Duemmel\\Desktop\\Projekt\\TimeTestModels\\BTHideDelegate\\gt\\BugTracker_createProject.gtEngine");
		dest = new File("C:\\Users\\Duemmel\\runtime-EclipseApplication2\\Bugtracker\\model\\gt\\BugTracker_createProject.gtEngine");
		inputChannel = null;
		outputChannel = null;
		try {
			inputChannel = new FileInputStream(source).getChannel();
			outputChannel = new FileOutputStream(dest).getChannel();
			outputChannel.transferFrom(inputChannel, 0, inputChannel.size());
		} catch (Throwable e) {	e.printStackTrace();} 
		finally {
			try {
				inputChannel.close();
				outputChannel.close();
			} catch (IOException e) {e.printStackTrace();}
		}
		System.out.println(i);
		i++;	
		EReference newRef = EcoreFactory.eINSTANCE.createEReference();
		sourceClass.getEStructuralFeatures().add(1, newRef);
		newRef.setEType(delegate);
		newRef.setName("trackedProjects");
		
		while(i<15){
			execute2(delegate);
		}
		return null;
	}

	//////////////////////////////////////////////////////////////////
	//	Complete Java Implementation for the sake of testing below	//
	//////////////////////////////////////////////////////////////////

	@SuppressWarnings("rawtypes")
	private void cleanUpAndSave2(EList<EObject> objectsToRemove, Resource res){
		TreeIterator<EObject> iter = res.getAllContents();
		EList<EObject> nacsToRemove = new BasicEList<EObject>();
		while (iter.hasNext()){
			EObject eObject = iter.next();
			if(eObject instanceof  GraphTransformationRule<?, ?>){
				for(Object nac : ((GraphTransformationRule)eObject).getNegativePattern()){
					if(((GTNegativeApplicationCondition)nac).getElements().isEmpty())
						{nacsToRemove.add((EObject) nac);}
				}
			}
		}
		if(objectsToRemove != null && !objectsToRemove.isEmpty()){
			for(EObject o : objectsToRemove){
				EcoreUtil.remove(o);
			}
		}
		if(!nacsToRemove.isEmpty()){
			for(EObject o : nacsToRemove){
				EcoreUtil.remove(o);
			}
		}
		try {
			res.save(java.util.Collections.EMPTY_MAP);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	@SuppressWarnings({ "rawtypes" })
	private void applyToAffectedObjects2(ENamedElement eNamedElement) {
		for (Resource res : gtResources) {
			TreeIterator<EObject> iter = res.getAllContents();
			EList<EObject> objectsToRemove = new BasicEList<EObject>();
			String ruleName = "";
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					ruleName = ((GraphTransformationRule) eObject).getName();
					GraphTransformationRule rule = (GraphTransformationRule) eObject;
					EcoreUtil.resolveAll(eObject);					
					EList<GPElement> patternElements = rule.getPattern().getElements();
					GTNode sourceNodePattern = null;
					for(GPElement gpe : patternElements){
						if(gpe instanceof GTThisObject && ((EClass) ((GTThisObject)gpe).getType()).getName().equals(sourceClass.getName()))
							sourceNodePattern = (GTNode) gpe;
						else if(gpe instanceof GPUnboundNode && ((EClass) ((GPUnboundNode)gpe).getType()).getName().equals(sourceClass.getName()))
							sourceNodePattern = (GTNode) gpe;
						else if(gpe instanceof GPBoundNode && ((EClass) ((GPBoundNode)gpe).getType()).getName().equals(sourceClass.getName()))
							sourceNodePattern = (GTNode) gpe;
					}
					if(sourceNodePattern != null){
						EList gtLinks = sourceNodePattern.getOutgoingEdges();
						for(int i=gtLinks.size()-1; i >= 0; i--){
							GTLink gta = (GTLink) gtLinks.get(i);
							if(gta.getExReference().eIsProxy()){
								objectsToRemove.add(gta);
							}
						}
					}
					for(Object nac : rule.getNegativePattern()){
						GTNegativeApplicationCondition temp = (GTNegativeApplicationCondition) nac;
						EList<NACElement> nacElements = temp.getElements();
						GTNode sourceNodeNAC = null;
						for(NACElement gpe : nacElements){
							if(gpe instanceof GTThisObject && ((EClass) ((GTThisObject)gpe).getType()).getName().equals(sourceClass.getName()))
								sourceNodeNAC = (GTNode) gpe;
							else if(gpe instanceof NACObject && ((EClass) ((NACObject)gpe).getType()).getName().equals(sourceClass.getName()))
								sourceNodeNAC = (GTNode) gpe;
							else if(gpe instanceof NACParameter && ((EClass) ((NACParameter)gpe).getType()).getName().equals(sourceClass.getName()))
								sourceNodeNAC = (GTNode) gpe;
							else if(gpe instanceof NACBoundNode && ((EClass) ((NACBoundNode)gpe).getGPUnboundNode().getType()).getName().equals(sourceClass.getName()))
								sourceNodeNAC = (GTNode) gpe;
						}
						if(sourceNodeNAC != null){
							EList gtLinks = sourceNodeNAC.getOutgoingEdges();
							for(int i=gtLinks.size()-1; i >= 0; i--){
								GTLink gta = (GTLink) gtLinks.get(i);
								if(gta.getExReference().eIsProxy()){
									objectsToRemove.add(gta);
								}
							}
						}
					}					
				} 
				else if (eObject instanceof GTField	&& ((GTField) eObject).getKind().getName().equals("OCL_CONSTRAINT")) {
					String containerName = "OCL_CONSTRAINT";
					String expression = ((GTField) eObject).getContent();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTField) eObject).setContent(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if (eObject instanceof GTPath) {
					String containerName = "GTPath";
					String expression = ((GTPath) eObject).getExpression();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTPath) eObject).setExpression(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPrecondition){
					String containerName = "GTPrecondition";
					String expression = ((GTPrecondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTPrecondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPostcondition){
					String containerName = "GTPostcondition";
					String expression = ((GTPostcondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTPostcondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof Shape && objectsToRemove.contains(((Shape)eObject).getElement()))
					objectsToRemove.add(eObject);	
				else if(eObject instanceof Edge && objectsToRemove.contains(((Edge)eObject).getElement()))
					objectsToRemove.add(eObject);				
			}
			cleanUpAndSave2(objectsToRemove, res);
//			try{
//				boolean result = callValidation(res);
//				if(result == false){
//					JOptionPane.showMessageDialog(null,
//							"Errors exist in the rule " + ruleName + "\nPlease check this rule.", "Validation failed!",
//							JOptionPane.ERROR_MESSAGE);
//				}
//			} catch (Throwable e){
//				e.printStackTrace();
//			}
		}
	}
	
//	@SuppressWarnings("unused")
//	private boolean propagateChanges2(ChangeDescription description,
//			ENamedElement eNamedElement) {
//		EList<FeatureChange> changes = description.getObjectChanges().get(
//				eNamedElement);
//		EList<EObject> addedObjects = description.getObjectsToAttach();
//		EList<EObject> removedObjects = description.getObjectsToDetach();
//		gtResources.clear();
//		getGTResources(eNamedElement);
//		if (changes != null)
//			for (FeatureChange change : changes) {
//				System.out.println(change + " Name: " + change.getFeatureName()
//						+ " Value: " + change.getDataValue());		
//			}
//		if (addedObjects != null)
//			for (EObject addedObj : addedObjects) {
//				System.out.println(addedObj);				
//			}
//		if (removedObjects != null)
//			for (EObject removedObj : removedObjects) {
//				System.out.println(removedObj);
//			}		
//		applyToAffectedObjects2(eNamedElement);
//		return false;
//	}
}