package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.io.IOException;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JOptionPane;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IHandler;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EParameter;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.presentation.EcoreEditor;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.gmf.runtime.notation.Edge;
import org.eclipse.gmf.runtime.notation.Node;
import org.eclipse.gmf.runtime.notation.Shape;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;

import de.ubt.ai1.modgraph.gt.gtEngine.GTField;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPath;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPostcondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPrecondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GraphTransformationRule;

public class InlineEClassConnector extends RefactoringConnector implements
		IHandler {
	private IEditorPart activeEditor;
	private EClass sourceClass;
	private String targetClassName;
	private EClass targetClass;

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		activeEditor = HandlerUtil.getActiveEditor(event);
		window = HandlerUtil.getActiveWorkbenchWindow(event);

		if (activeEditor instanceof EcoreEditor) {
			EcoreEditor ecoreEditor = (EcoreEditor) activeEditor;
			ecoreEditor.doSave(null);
			TreeSelection sel = (TreeSelection) ecoreEditor.getSelection();
			Object obj = sel.getFirstElement();

			if (obj instanceof EClass) {
				// parameter fuer die Refactoring Operation sammeln
				sourceClass = (EClass) obj;
				EPackage pack = sourceClass.getEPackage();
				Resource res = null;
				res = pack.eResource();
				List<String> classNames = new LinkedList<String>();
				List<EClassifier> classes = pack.getEClassifiers();
				for (EClassifier c : classes) {
					if (c instanceof EClass)
						classNames.add(c.getName());
				}
				classNames.remove(sourceClass.getName());
				targetClassName = getSelectedFromDialog(classNames,
						"Choose target class.");

				// refactoring und helper Objekte initialisieren
				initRefactoring(sourceClass.getEPackage().eContents());
				targetClass = helper.getClass(targetClassName);
				ChangeRecorder recorder = new ChangeRecorder(targetClass);
				try {
					for (EReference r : targetClass.getEReferences()) {
						if (r.getEType().equals(sourceClass))
							targetClass.getEStructuralFeatures().remove(r);
					}
					refactoring.inlineEClass(sourceClass.getName(),
							targetClassName);
					// inlined Class remains in the helper and needs to be
					// removed, so that updateOperationCalls() will work
					// properly
					EList<EObject> temp = new BasicEList<EObject>();
					for (EClass eClass : helper.getAllClasses())
						for (EOperation eOp : eClass.getEOperations())
							for (EParameter eParam : eOp.getEParameters()) {
								if (eParam.getEType().eIsProxy()
										|| eParam.getEType() == null)
									temp.add(eParam);
								else if (eParam.getEType().getName()
										.equals(sourceClass.getName())) {
									temp.add(eParam);
									temp.add(sourceClass);
								}
							}
					for (int i = temp.size() - 1; i >= 0; i--)
						EcoreUtil.delete(temp.get(i));
				} catch (Throwable e) {
					// JOptionPane.showMessageDialog(null,
					// "Refactoring failed.",
					// "ERROR!", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
					return null;
				}

				try {
					res.save(Collections.EMPTY_MAP);
				} catch (IOException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
				
				try {
					ChangeDescription description = recorder.endRecording();
					if (!(description.eContents().isEmpty())) {
						propagateChanges(description, targetClass,res);
					}
				} catch (Throwable e) {
					// JOptionPane.showMessageDialog(null,
					// "Propagation of refactoring changes to GTRules failed.",
					// "ERROR!", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
				}
			}

		}
		return null;
	}

	@SuppressWarnings({ "rawtypes" })
	@Override
	protected void applyToAffectedObjects(ENamedElement eNamedElement) {
		EList<Object> temp = new BasicEList<Object>();
		for (Resource res : gtResources) {
			temp.add((Object) res);
			EcoreUtil.resolveAll(res);
			TreeIterator<EObject> iter = res.getAllContents();
			EList<EObject> objectsToRemove = new BasicEList<EObject>();
			String ruleName = "";
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					ruleName = ((GraphTransformationRule) eObject).getName();
					objectsToRemove = propagation
							.propagateInlineClass_CollapseHierarchy(
									sourceClass, targetClass,
									(GraphTransformationRule) eObject,
									objectsToRemove);

				} 
				else if (eObject instanceof Node) {
					if (objectsToRemove.contains(((Node) eObject).getElement()))
						objectsToRemove.add(eObject);
				} 
				else if (eObject instanceof Edge) {
					if (objectsToRemove.contains(((Edge) eObject).getElement()))
						objectsToRemove.add(eObject);
				}

				else if (eObject instanceof GTField	&& ((GTField) eObject).getKind().getName().equals("OCL_CONSTRAINT")) {
					String containerName = "OCL_CONSTRAINT";
					String expression = ((GTField) eObject).getContent();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, sourceClass.getName(), containerName);
					if(newExpression != null)
						((GTField) eObject).setContent(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if (eObject instanceof GTPath) {
					String containerName = "GTPath";
					String expression = ((GTPath) eObject).getExpression();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, sourceClass.getName(), containerName);
					if(newExpression != null)
						((GTPath) eObject).setExpression(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPrecondition){
					String containerName = "GTPrecondition";
					String expression = ((GTPrecondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, sourceClass.getName(), containerName);
					if(newExpression != null)
						((GTPrecondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPostcondition){
					String containerName = "GTPostcondition";
					String expression = ((GTPostcondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, sourceClass.getName(), containerName);
					if(newExpression != null)
						((GTPostcondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof Shape && objectsToRemove.contains(((Shape)eObject).getElement()))
					objectsToRemove.add(eObject);
			}
			propagation.cleanUpAndSave(objectsToRemove, res);
		}
		propagation.updateOperationCalls(temp);
		for (Resource res : gtResources) {
			propagation.cleanUpAndSave(null, res);
			
			URI resURI = res.getURI();
			String fragment = resURI.toString();
			String[] resRuleNameArray = fragment.split("/");
			String resRuleName = resRuleNameArray[resRuleNameArray.length-1];			
			int response = JOptionPane.showConfirmDialog(null,
					"Would you like to validate the GT rule " + resRuleName + "?", "Attention",
					JOptionPane.YES_NO_OPTION);
			if(response == 0){
				try{
					boolean result = callValidation(res);
					if(result == false){
						JOptionPane.showMessageDialog(null,
								"Errors exist in the rule " + resRuleName + "\nPlease check it manually.", "Validation failed!",
								JOptionPane.ERROR_MESSAGE);
					}
				} catch (Throwable e){
					e.printStackTrace();
				}
			}
		}
	}

	// protected boolean propagateChanges2(ChangeDescription description,
	// ENamedElement eNamedElement) {
	// EList<FeatureChange> changes = description.getObjectChanges().get(
	// eNamedElement);
	// EList<EObject> addedObjects = description.getObjectsToAttach();
	// EList<EObject> removedObjects = description.getObjectsToDetach();
	// getGTResources(eNamedElement);
	// if (changes != null)
	// for (FeatureChange change : changes) {
	// System.out.println(change + " Name: " + change.getFeatureName()
	// + " Value: " + change.getDataValue());
	// applyToAffectedObjects2(change, eNamedElement);
	// }
	// if (addedObjects != null)
	// for (EObject addedObj : addedObjects) {
	// System.out.println(addedObj);
	//
	// }
	// if (removedObjects != null)
	// for (EObject removedObj : removedObjects) {
	// System.out.println(removedObj);
	// }
	// return false;
	// }
	//
	// /**
	// * Moves attributes and fields. Reroutes or deletes links if necessary.
	// * Removes the corresponding node.
	// */
	// @SuppressWarnings({ "unchecked", "rawtypes" })
	// protected List<EObject> applyToAffectedObjects2(FeatureChange change,
	// ENamedElement eNamedElement) {
	// // ValidationMenuHandler vmh = new ValidationMenuHandler();
	// for (Resource res : gtResources) {
	// TreeIterator<EObject> iter = res.getAllContents();
	// EList<EObject> objectsToRemove = new BasicEList<EObject>();
	// while (iter.hasNext()) {
	// EObject eObject = iter.next();
	// if (eObject instanceof GraphTransformationRule<?, ?>) {
	// if (((GraphTransformationRule) eObject)
	// .getImplementedOperation().eIsProxy()) {
	// String implOpName = propagation
	// .getNameFromURI(((GraphTransformationRule) eObject)
	// .getImplementedOperation());
	// for (EOperation o : targetClass.getEOperations()) {
	// if (o.getName().equals(implOpName)) {
	// ((GraphTransformationRule) eObject)
	// .setCorrespondingClass(targetClass);
	// ((GraphTransformationRule) eObject)
	// .setImplementedOperation(o);
	// }
	// }
	// }
	// GTGraphPattern pattern = ((GraphTransformationRule) eObject)
	// .getPattern();
	// GPElement sourceNode = null;
	// GPElement targetNode = null;
	// for (GPElement gtNode : pattern.getElements()) {
	// if (gtNode instanceof GPUnboundNode
	// && ((GPUnboundNode) gtNode).getType()
	// .eIsProxy())
	// sourceNode = gtNode;
	// else if (gtNode instanceof GTThisObject
	// && ((GTThisObject) gtNode).getType().eIsProxy())
	// sourceNode = gtNode;
	// else if (gtNode instanceof GPUnboundNode
	// && ((EClass) ((GPUnboundNode) gtNode).getType())
	// .getName().equals(targetClassName))
	// targetNode = gtNode;
	// else if (gtNode instanceof GTThisObject
	// && ((EClass) ((GTThisObject) gtNode).getType())
	// .getName().equals(targetClassName))
	// targetNode = gtNode;
	// }
	// moveElements((GTNode) sourceNode, (GTNode) targetNode,
	// objectsToRemove);
	//
	// List<GTNegativeApplicationCondition> nacList = ((GraphTransformationRule)
	// eObject)
	// .getNegativePattern();
	// for (GTNegativeApplicationCondition nac : nacList) {
	// NACElement sourceNodeNAC = null;
	// NACElement targetNodeNAC = null;
	// for (NACElement nacNode : nac.getElements()) {
	// if (nacNode instanceof NACBoundNode
	// && ((NACBoundNode) nacNode)
	// .getGPUnboundNode().getType()
	// .eIsProxy())
	// sourceNodeNAC = nacNode;
	// else if (nacNode instanceof NACObject
	// && ((NACObject) nacNode).getType()
	// .eIsProxy())
	// sourceNodeNAC = nacNode;
	// else if (nacNode instanceof GTThisObject
	// && ((GTThisObject) nacNode).getType()
	// .eIsProxy())
	// sourceNodeNAC = nacNode;
	// else if (nacNode instanceof NACBoundNode
	// && ((EClass) ((NACBoundNode) nacNode)
	// .getGPUnboundNode().getType())
	// .getName().equals(targetClassName))
	// targetNodeNAC = nacNode;
	// else if (nacNode instanceof NACObject
	// && ((EClass) ((NACObject) nacNode)
	// .getType()).getName().equals(
	// targetClassName))
	// targetNodeNAC = nacNode;
	// else if (nacNode instanceof GTThisObject
	// && ((EClass) ((GTThisObject) nacNode)
	// .getType()).getName().equals(
	// targetClassName))
	// targetNodeNAC = nacNode;
	// }
	// // if targetNode is null, retype the sourceNode to
	// // targetClass
	// if (sourceNodeNAC != null && targetNodeNAC != null)
	// moveElements((GTNode) sourceNodeNAC,
	// (GTNode) targetNodeNAC, objectsToRemove);
	// else if (sourceNodeNAC != null && targetNodeNAC == null) {
	// if (sourceNodeNAC instanceof NACBoundNode) {
	// if (targetNode instanceof GPUnboundNode) {
	// ((NACBoundNode) sourceNodeNAC)
	// .setGPUnboundNode((GPUnboundNode) targetNode);
	// retypeElements(sourceNodeNAC,
	// objectsToRemove);
	// } else if (targetNode instanceof GTThisObject)
	// objectsToRemove.add(sourceNodeNAC);
	// } else if (sourceNodeNAC instanceof NACObject) {
	// ((NACObject) sourceNodeNAC)
	// .setType(targetClass);
	// retypeElements(sourceNodeNAC, objectsToRemove);
	// } else if (sourceNodeNAC instanceof GTThisObject) {
	// ((GTThisObject) sourceNodeNAC)
	// .setType(targetClass);
	// retypeElements(sourceNodeNAC, objectsToRemove);
	// }
	// }
	// }
	// } else if (eObject instanceof Node)
	// if (objectsToRemove.contains(((Node) eObject).getElement()))
	// objectsToRemove.add(eObject);
	// else if (eObject instanceof Edge)
	// if (objectsToRemove.contains(((Edge) eObject)
	// .getElement()))
	// objectsToRemove.add(eObject);
	// }
	// propagation.cleanUpAndSave(objectsToRemove, res);
	// // vmh.initValidation(gtFiles.get(gtResources.indexOf(res)), window,
	// // new NullProgressMonitor());
	// }
	// return null;
	// }
	//
	// @SuppressWarnings({ "unchecked", "rawtypes" })
	// private void retypeElements(NACElement sourceNodeNAC,
	// List<EObject> objectsToRemove) {
	// for (GTAttribute gta : ((GTNode) sourceNodeNAC).getAttributes())
	// if (gta.getExAttribute().eIsProxy()) {
	// String gtaName = propagation.getNameFromURI(gta
	// .getExAttribute());
	// for (EAttribute ea : targetClass.getEAttributes())
	// if (ea.getName().equals(gtaName))
	// gta.setExAttribute(ea);
	// }
	// for (GTEdge gtl : ((GTNode) sourceNodeNAC).getIncomingEdges())
	// if (gtl.getSourceNode().eIsProxy()
	// || gtl.getTargetNode().eIsProxy())
	// objectsToRemove.add(gtl);
	// for (GTEdge gtl : ((GTNode) sourceNodeNAC).getOutgoingEdges())
	// if (gtl.getSourceNode().eIsProxy()
	// || gtl.getTargetNode().eIsProxy())
	// objectsToRemove.add(gtl);
	// }
	//
	// // For some reason does not work with lists (does not stop at the end -->
	// // NullPointerException) so I had to use arrays.
	// @SuppressWarnings("unchecked")
	// private void moveElements(GTNode sourceNode, GTNode targetNode,
	// List<EObject> objectsToRemove) {
	// if (sourceNode != null && targetNode == null) {
	// objectsToRemove.add(sourceNode);
	// objectsToRemove.addAll(((GTNode) sourceNode).getAttributes());
	// objectsToRemove.addAll(((GTNode) sourceNode).getFields());
	// objectsToRemove.addAll(((GTNode) sourceNode).getIncomingEdges());
	// objectsToRemove.addAll(((GTNode) sourceNode).getOutgoingEdges());
	// } else if (sourceNode != null && targetNode != null) {
	// objectsToRemove.add(sourceNode);
	// EList<GTAttribute<EAttribute>> attributes =
	// (EList<GTAttribute<EAttribute>>) (EList<?>) (((GTNode) sourceNode)
	// .getAttributes());
	// int size = attributes.size();
	// for (int i = size - 1; i >= 0; i--) {
	// GTAttribute<EAttribute> a = attributes.get(i);
	// ((GTNode) targetNode).getAttributes().add(a);
	// if (targetNode instanceof GTThisObject
	// && targetNode.eContainer() instanceof GTGraphPattern)
	// moveShape(targetNode.eResource(), a, targetNode, "2001");
	// else if (targetNode instanceof GPObject)
	// moveShape(targetNode.eResource(), a, targetNode, "2011");
	// else if (targetNode instanceof NACObject)
	// moveShape(targetNode.eResource(), a, targetNode, "2017");
	// else if (targetNode instanceof GTThisObject
	// && targetNode.eContainer() instanceof GTNegativeApplicationCondition)
	// moveShape(targetNode.eResource(), a, targetNode, "2018");
	// EAttribute exAttr = a.getExAttribute();
	// if (exAttr.eIsProxy()) {
	// String exAttrName = propagation.getNameFromURI(exAttr);
	// for (EAttribute eAttr : targetClass.getEAttributes())
	// if (eAttr.getName().equals(exAttrName))
	// a.setExAttribute(eAttr);
	// }
	// }
	// // ((GTNode) targetNode).getFields().addAll(((GTNode)
	// // sourceNode).getFields());
	// EList<GTField> fields = sourceNode.getFields();
	// size = fields.size();
	// for (int i = size - 1; i >= 0; i--) {
	// GTField f = fields.get(i);
	// targetNode.getFields().add(f);
	// if (targetNode instanceof GTThisObject
	// && targetNode.eContainer() instanceof GTGraphPattern)
	// moveShape(targetNode.eResource(), f, targetNode, "2001");
	// else if (targetNode instanceof GPObject)
	// moveShape(targetNode.eResource(), f, targetNode, "2011");
	// else if (targetNode instanceof NACObject)
	// moveShape(targetNode.eResource(), f, targetNode, "2017");
	// else if (targetNode instanceof GTThisObject
	// && targetNode.eContainer() instanceof GTNegativeApplicationCondition)
	// moveShape(targetNode.eResource(), f, targetNode, "2018");
	// }
	// EList<GTLink<EReference>> incomingLinks = (EList<GTLink<EReference>>)
	// (EList<?>) (((GTNode) sourceNode)
	// .getIncomingEdges());
	// size = incomingLinks.size();
	// for (int i = size - 1; i >= 0; i--) {
	// GTLink<EReference> l = incomingLinks.get(i);
	// objectsToRemove.add(l);
	// }
	// EList<GTLink<EReference>> outgoingLinks = (EList<GTLink<EReference>>)
	// (EList<?>) (((GTNode) sourceNode)
	// .getOutgoingEdges());
	// size = outgoingLinks.size();
	// for (int i = size - 1; i >= 0; i--) {
	// GTLink<EReference> l = outgoingLinks.get(i);
	// l.setSourceNode((GTNode) targetNode);
	// EReference exRef = l.getExReference();
	// if (exRef.eIsProxy()) {
	// String exRefName = propagation.getNameFromURI(exRef);
	// for (EReference r : targetClass.getEReferences())
	// if (r.getName().equals(exRefName))
	// l.setExReference(r);
	// }
	// }
	// }
	// }
}