package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.channels.FileChannel;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Pattern;

import javax.swing.JOptionPane;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.FeatureChange;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.presentation.EcoreEditor;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.gmf.runtime.notation.BasicCompartment;
import org.eclipse.gmf.runtime.notation.Node;
import org.eclipse.gmf.runtime.notation.Shape;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.ui.handlers.HandlerUtil;

import de.ubt.ai1.modGraph.example.eRefactoring.ERefactoringFactory;
import de.ubt.ai1.modgraph.gt.gtEngine.GPObject;
import de.ubt.ai1.modgraph.gt.gtEngine.GTAttribute;
import de.ubt.ai1.modgraph.gt.gtEngine.GTField;
import de.ubt.ai1.modgraph.gt.gtEngine.GTNegativeApplicationCondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPath;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPostcondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPrecondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTThisObject;
import de.ubt.ai1.modgraph.gt.gtEngine.GraphTransformationRule;

public class MoveEOperationConnector extends RefactoringConnector {

	private String opName = "";
	private EClass targetClass;
	private EClass sourceClass;
	private int i = 0;

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		activeEditor = HandlerUtil.getActiveEditor(event);
		window = HandlerUtil.getActiveWorkbenchWindow(event);

		if (activeEditor instanceof EcoreEditor) {
			EcoreEditor ecoreEditor = (EcoreEditor) activeEditor;
			ecoreEditor.doSave(null);

			TreeSelection sel = (TreeSelection) ecoreEditor.getSelection();
			Object obj = sel.getFirstElement();

			if (obj instanceof EOperation) {
				EOperation op = (EOperation) obj;
				opName = op.getName();

				sourceClass = (EClass) op.eContainer();
				ChangeRecorder recorder = new ChangeRecorder(op);
				EPackage pack = sourceClass.getEPackage();
				List<String> classNames = new LinkedList<String>();
				List<EClassifier> classes = pack.getEClassifiers();
				for (EClassifier c : classes) {
					if (c instanceof EClass)
						classNames.add(c.getName());
				}
				String targetClassName = getSelectedFromDialog(classNames,
						"Choose target class.");

				initRefactoring(sourceClass.getEPackage().eContents());
				targetClass = helper.getClass(targetClassName);
				
				try {
					refactoring.moveEOperation(op.getName(), targetClassName,	sourceClass.getName());
				} catch (Throwable e) {
					JOptionPane.showMessageDialog(
									null,
									"The target class already contains an operation with the same name.",
									"ERROR", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
					return null;
				}
				
				try {
					sourceClass.getEPackage().eResource().save(Collections.EMPTY_MAP);
				} catch (IOException e) {e.printStackTrace();}				
				
				ChangeDescription description = recorder.endRecording();
				if (!(description.eContents().isEmpty())) {
					propagateChanges(description, op, sourceClass.eResource());
				}			
				
				
//				// For runtime testing only
//				File source = new File("C:\\Users\\Duemmel\\Desktop\\Projekt\\TimeTestModels\\BTMoveAttribute\\BugModel.ecore");
//				File dest = new File("C:\\Users\\Duemmel\\runtime-EclipseApplication2\\Bugtracker\\model\\BugModel.ecore");
//				FileChannel inputChannel = null;
//				FileChannel outputChannel = null;
//				try {
//					inputChannel = new FileInputStream(source).getChannel();
//					outputChannel = new FileOutputStream(dest).getChannel();
//					outputChannel.transferFrom(inputChannel, 0, inputChannel.size());
//				} catch (Throwable e) {	e.printStackTrace();} 
//				finally {
//					try {
//						inputChannel.close();
//						outputChannel.close();
//					} catch (IOException e) {
//						e.printStackTrace();
//					}
//				}
//				source = new File("C:\\Users\\Duemmel\\Desktop\\Projekt\\TimeTestModels\\BTMoveAttribute\\gt\\Project_createTicket.gtEngine");
//				dest = new File("C:\\Users\\Duemmel\\runtime-EclipseApplication2\\Bugtracker\\model\\gt\\Project_createTicket.gtEngine");
//				inputChannel = null;
//				outputChannel = null;
//				try {
//					inputChannel = new FileInputStream(source).getChannel();
//					outputChannel = new FileOutputStream(dest).getChannel();
//					outputChannel.transferFrom(inputChannel, 0, inputChannel.size());
//				} catch (Throwable e) {	e.printStackTrace();} 
//				finally {
//					try {
//						inputChannel.close();
//						outputChannel.close();
//					} catch (IOException e) {e.printStackTrace();}
//				}
//				execute2();
//				i = 0;			
			}
		}
		return null;
	}

	/**
	 * Changes the correspondingClass of GTRules whose implementedOperation was
	 * moved to another class. If the targetClass is in the rule moves the
	 * OPERATION_CALL to it, else deletes the OPERATION_CALL from its containing
	 * class.
	 */
	@SuppressWarnings({ "rawtypes" })
	@Override
	protected void applyToAffectedObjects(ENamedElement eNamedElement) {
		for (Resource res : gtResources) {
			TreeIterator<EObject> iter = res.getAllContents();
			EList<EObject> objectsToRemove = new BasicEList<EObject>();
			String ruleName = "";
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					ruleName = ((GraphTransformationRule) eObject).getName();
					objectsToRemove = propagation.propagateMoveOperation((EOperation) eNamedElement, sourceClass, targetClass, (GraphTransformationRule) eObject, objectsToRemove);//						
				} else if (eObject instanceof Shape
						&& objectsToRemove.contains(((Shape) eObject)
								.getElement())) {
					objectsToRemove.add(eObject);
				} 
				else if (eObject instanceof GTField	&& ((GTField) eObject).getKind().getName().equals("OCL_CONSTRAINT")) {
					String containerName = "OCL_CONSTRAINT";
					String expression = ((GTField) eObject).getContent();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTField) eObject).setContent(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if (eObject instanceof GTPath) {
					String containerName = "GTPath";
					String expression = ((GTPath) eObject).getExpression();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTPath) eObject).setExpression(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPrecondition){
					String containerName = "GTPrecondition";
					String expression = ((GTPrecondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTPrecondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPostcondition){
					String containerName = "GTPostcondition";
					String expression = ((GTPostcondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTPostcondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof Shape && objectsToRemove.contains(((Shape)eObject).getElement()))
					objectsToRemove.add(eObject);
			}
			propagation.cleanUpAndSave(objectsToRemove, res);
			
			URI resURI = res.getURI();
			String fragment = resURI.toString();
			String[] resRuleNameArray = fragment.split("/");
			String resRuleName = resRuleNameArray[resRuleNameArray.length-1];			
			int response = JOptionPane.showConfirmDialog(null,
					"Would you like to validate the GT rule " + resRuleName + "?", "Attention",
					JOptionPane.YES_NO_OPTION);
			if(response == 0){
				try{
					boolean result = callValidation(res);
					if(result == false){
						JOptionPane.showMessageDialog(null,
								"Errors exist in the rule " + resRuleName + "\nPlease check it manually.", "Validation failed!",
								JOptionPane.ERROR_MESSAGE);
					}
				} catch (Throwable e){
					e.printStackTrace();
				}
			}
		}
	}
	
	
	// for runtime testing
	public Object execute2() throws ExecutionException {
		EcoreUtil.resolveAll(sourceClass.eResource());
		EOperation op = helper.getOperationByName(targetClass, opName);
		sourceClass.getEOperations().add(3, op);
		helper.removeOperationByName(targetClass, opName);
		ChangeRecorder recorder = new ChangeRecorder(op);
		refactoring = ERefactoringFactory.eINSTANCE.createRefactoring();
		helper = ERefactoringFactory.eINSTANCE.createHelper();
		propagation = ERefactoringFactory.eINSTANCE.createPropagation();
		initRefactoring(sourceClass.getEPackage().eContents());		
		String targetClassName = targetClass.getName();
		
		long preRefactoringNano = System.nanoTime();		
		try {
			refactoring.moveEOperation(op.getName(), targetClassName, sourceClass.getName());
		} catch (Throwable e) {
			JOptionPane.showMessageDialog(
							null,
							"The target class already contains an operation with the same name.",
							"ERROR", JOptionPane.ERROR_MESSAGE);
			e.printStackTrace();
			return null;
		}
		
		
//		targetClass.getEOperations().add(op);  //for runtime testing purpose only
		
		try {
			sourceClass.getEPackage().eResource().save(Collections.EMPTY_MAP);
		} catch (IOException e) {e.printStackTrace();}
		
		long prePropagationNano = System.nanoTime();
		
		ChangeDescription description = recorder.endRecording();
		if (!(description.eContents().isEmpty())) {
			propagateChanges(description, op, op.eResource());
		}
		
//		if (!(description.eContents().isEmpty())) {
//			propagateChanges2(description, op);
//		}		
		
		long finalNano = System.nanoTime();
		
		try(PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter("C:/Move_EOperation_Refactoring.txt", true)))) {
			out.println((prePropagationNano - preRefactoringNano));
		}catch (IOException e) {}			
		try(PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter("C:/Move_EOperation_Propagation.txt", true)))) {
			out.println((finalNano - prePropagationNano));
		}catch (IOException e) {}
		
		// For runtime testing only
		File source = new File("C:\\Users\\Duemmel\\Desktop\\Projekt\\TimeTestModels\\BTMoveAttribute\\BugModel.ecore");
		File dest = new File("C:\\Users\\Duemmel\\runtime-EclipseApplication2\\Bugtracker\\model\\BugModel.ecore");
		FileChannel inputChannel = null;
		FileChannel outputChannel = null;
		try {
			inputChannel = new FileInputStream(source).getChannel();
			outputChannel = new FileOutputStream(dest).getChannel();
			outputChannel.transferFrom(inputChannel, 0, inputChannel.size());
		} catch (Throwable e) {	e.printStackTrace();} 
		finally {
			try {
				inputChannel.close();
				outputChannel.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		source = new File("C:\\Users\\Duemmel\\Desktop\\Projekt\\TimeTestModels\\BTMoveAttribute\\gt\\Project_createTicket.gtEngine");
		dest = new File("C:\\Users\\Duemmel\\runtime-EclipseApplication2\\Bugtracker\\model\\gt\\Project_createTicket.gtEngine");
		inputChannel = null;
		outputChannel = null;
		try {
			inputChannel = new FileInputStream(source).getChannel();
			outputChannel = new FileOutputStream(dest).getChannel();
			outputChannel.transferFrom(inputChannel, 0, inputChannel.size());
		} catch (Throwable e) {	e.printStackTrace();} 
		finally {
			try {
				inputChannel.close();
				outputChannel.close();
			} catch (IOException e) {e.printStackTrace();}
		}
		System.out.println(i);
		i++;
		while(i<15){
			execute2();
		}
		return null;

	}

	//////////////////////////////////////////////////////////////////
	//	Complete Java Implementation for the sake of testing below	//
	//////////////////////////////////////////////////////////////////

	@SuppressWarnings("rawtypes")
	private void cleanUpAndSave2(EList<EObject> objectsToRemove, Resource res){
		TreeIterator<EObject> iter = res.getAllContents();
		EList<EObject> nacsToRemove = new BasicEList<EObject>();
		while (iter.hasNext()){
			EObject eObject = iter.next();
			if(eObject instanceof  GraphTransformationRule<?, ?>){
				for(Object nac : ((GraphTransformationRule)eObject).getNegativePattern()){
					if(((GTNegativeApplicationCondition)nac).getElements().isEmpty())
						{nacsToRemove.add((EObject) nac);}
				}
			}
		}
		if(objectsToRemove != null && !objectsToRemove.isEmpty()){
			for(EObject o : objectsToRemove){
				EcoreUtil.remove(o);
			}
		}
		if(!nacsToRemove.isEmpty()){
			for(EObject o : nacsToRemove){
				EcoreUtil.remove(o);
			}
		}
		try {
			res.save(java.util.Collections.EMPTY_MAP);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	@SuppressWarnings({ "rawtypes", "unchecked" })
	private void moveShape2(EObject movedObject, EObject newContainer, String nodeType, GraphTransformationRule rule) {
		TreeIterator<EObject> iter2 = rule.eResource().getAllContents();
		Shape shape = null;
		Node targetNode = null;
		while(iter2.hasNext()){
			EObject next = iter2.next();
			if(next instanceof Shape && ((Shape) next).getElement().equals(movedObject))
				{shape = ((Shape) next);}
			if(next instanceof Node && ((Node) next).getType().equals(nodeType) && ((Node) next).getElement().equals(newContainer))
				{targetNode = ((Node) next);}
		}
		if(shape != null && targetNode != null){
			if(shape.eContainer().eContainer().eContainer().equals(targetNode.eContainer())){	//important when multiple NACs are used in one rule
				for(Object o : targetNode.getPersistedChildren()){
					if(o instanceof BasicCompartment 
						&& ((shape.getElement() instanceof GTAttribute && !((GTAttribute) shape.getElement()).getOperator().getLiteral().equals("="))
						|| (shape.getElement() instanceof GTField && ((GTField) shape.getElement()).getKind().getName().equals("OCL_CONSTRAINT")))
						){
						((BasicCompartment) o).getPersistedChildren().add(shape);
						break;
					}
					else if(o instanceof BasicCompartment
							&& ((shape.getElement() instanceof GTAttribute && ((GTAttribute) shape.getElement()).getOperator().getLiteral().equals("=")) 
							|| (shape.getElement() instanceof GTField && ((GTField) shape.getElement()).getKind().getName().equals("OPERATION_CALL")))
							)
						{((BasicCompartment) o).getPersistedChildren().add(shape);}
				}
			}
		}
	}
	
	@SuppressWarnings({ "rawtypes", "unchecked" })
	private void applyToAffectedObjects2(ENamedElement eNamedElement) {
		for (Resource res : gtResources) {
			TreeIterator<EObject> iter = res.getAllContents();
			EList<EObject> objectsToRemove = new BasicEList<EObject>();
			String ruleName = "";
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					ruleName = ((GraphTransformationRule) eObject).getName();
					GraphTransformationRule rule = (GraphTransformationRule) eObject;
					EcoreUtil.resolveAll(eObject);
					if(rule.getImplementedOperation().eIsProxy()){
						EOperation implOp = (EOperation) rule.getImplementedOperation();
						if(implOp.eIsProxy()){
							String fragment = EcoreUtil.getURI(implOp).fragment();
							String[] implOpNameArray = fragment.split("/");		
							String implOpName = implOpNameArray[implOpNameArray.length-1];
							for(EOperation eOp : targetClass.getEOperations()){
								if(eOp.getName().equals(implOpName)){
									rule.setCorrespondingClass(targetClass);
									rule.setImplementedOperation(eOp);
								}
							}
						}	
					}
					TreeIterator<EObject> iter2 = rule.eAllContents();
					boolean moved = false;
					while(iter2.hasNext()){
						if(moved)
							break;
						EObject eObject2 = iter2.next();		 
						if(eObject2 instanceof GTField && ((GTField) eObject2).getKind().getName().equals("OPERATION_CALL")){
							String operationName = ((GTField) eObject2).getContent().split(Pattern.quote("("))[0];
							if (operationName.equals(((EOperation) eNamedElement).getName())) {		
								for(EObject ruleObject : rule.getPattern().getElements()){
									if (ruleObject instanceof GPObject && !moved) {
										EClass eClass = (EClass) ((GPObject)ruleObject).getType();
										String name = eClass.getName();
										if (name.equals(targetClass.getName())) {
											((GPObject)ruleObject).getFields().add((GTField) eObject2);
											String status = ((GPObject)ruleObject).getStatus().getName();
											if(status.equals("CREATE"))
												{moveShape2(eObject, ruleObject, "2013", rule);}
											else if(status.equals("DELETE"))
												{moveShape2(eObject, ruleObject, "2012", rule);}
											else
												{moveShape2(eObject, ruleObject, "2011", rule);}
											moved = true;
										}
									} else if (ruleObject instanceof GTThisObject && !moved) {
										EClass eClass = (EClass) ((GTThisObject)ruleObject).getType();
										String name = eClass.getName();
										if (name.equals(targetClass.getName())) {
											((GTThisObject)ruleObject).getFields().add((GTField) eObject2);
											moveShape2(eObject2, ruleObject, "2001", rule);
											moved = true;
										}
									}
								}
								if (!moved) {
									objectsToRemove.add(eObject);
									moved = true;
								}
							}
						}
					}
				} 
				else if (eObject instanceof GTField	&& ((GTField) eObject).getKind().getName().equals("OCL_CONSTRAINT")) {
					String containerName = "OCL_CONSTRAINT";
					String expression = ((GTField) eObject).getContent();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTField) eObject).setContent(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if (eObject instanceof GTPath) {
					String containerName = "GTPath";
					String expression = ((GTPath) eObject).getExpression();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTPath) eObject).setExpression(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPrecondition){
					String containerName = "GTPrecondition";
					String expression = ((GTPrecondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTPrecondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPostcondition){
					String containerName = "GTPostcondition";
					String expression = ((GTPostcondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTPostcondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof Shape && objectsToRemove.contains(((Shape)eObject).getElement()))
					objectsToRemove.add(eObject);
			}
			cleanUpAndSave2(objectsToRemove, res);
//			try{
//				boolean result = callValidation(res);
//				if(result == false){
//					JOptionPane.showMessageDialog(null,
//							"Errors exist in the rule " + ruleName + "\nPlease check this rule.", "Validation failed!",
//							JOptionPane.ERROR_MESSAGE);
//				}
//			} catch (Throwable e){
//				e.printStackTrace();
//			}
		}
	}
	
	@SuppressWarnings("unused")
	private boolean propagateChanges2(ChangeDescription description,
			ENamedElement eNamedElement) {
		EList<FeatureChange> changes = description.getObjectChanges().get(
				eNamedElement);
		EList<EObject> addedObjects = description.getObjectsToAttach();
		EList<EObject> removedObjects = description.getObjectsToDetach();
		gtResources.clear();
		getGTResources(eNamedElement.eResource());
		if (changes != null)
			for (FeatureChange change : changes) {
				System.out.println(change + " Name: " + change.getFeatureName()
						+ " Value: " + change.getDataValue());		
			}
		if (addedObjects != null)
			for (EObject addedObj : addedObjects) {
				System.out.println(addedObj);				
			}
		if (removedObjects != null)
			for (EObject removedObj : removedObjects) {
				System.out.println(removedObj);
			}		
		applyToAffectedObjects2(eNamedElement);
		return false;
	}
}
