package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.swing.JOptionPane;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.presentation.EcoreEditor;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.gmf.runtime.notation.Shape;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;

import de.ubt.ai1.modgraph.gt.failure.GTFailure;
import de.ubt.ai1.modgraph.gt.gtEngine.GPBoundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GPElement;
import de.ubt.ai1.modgraph.gt.gtEngine.GPUnboundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GTAttribute;
import de.ubt.ai1.modgraph.gt.gtEngine.GTField;
import de.ubt.ai1.modgraph.gt.gtEngine.GTNegativeApplicationCondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPath;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPostcondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPrecondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTThisObject;
import de.ubt.ai1.modgraph.gt.gtEngine.GraphTransformationRule;
import de.ubt.ai1.modgraph.gt.gtEngine.NACBoundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.NACElement;
import de.ubt.ai1.modgraph.gt.gtEngine.NACObject;

public class PushDownEAttributeConnector extends RefactoringConnector {
	private IEditorPart activeEditor;
	private EClass sourceClass;
	private EAttribute attr;
	private String attrName;

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		activeEditor = HandlerUtil.getActiveEditor(event);
		window = HandlerUtil.getActiveWorkbenchWindow(event);

		if (activeEditor instanceof EcoreEditor) {
			EcoreEditor ecoreEditor = (EcoreEditor) activeEditor;
			ecoreEditor.doSave(null);
			TreeSelection sel = (TreeSelection) ecoreEditor.getSelection();
			Object obj = sel.getFirstElement();

			if (obj instanceof EAttribute) {
				attr = (EAttribute) obj;
				attrName = attr.getName();
				sourceClass = (EClass) attr.eContainer();
				ChangeRecorder recorder = new ChangeRecorder(sourceClass);

				initRefactoring(sourceClass.getEPackage().eContents());
				try {
					refactoring.pushDownEAttribute(attr.getName(),
							sourceClass.getName());
				} catch (Throwable e) {
					JOptionPane.showMessageDialog(null, "Refactoring failed.",
							"ERROR!", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
					return null;
				}
				try {
					sourceClass.getEPackage().eResource().save(Collections.EMPTY_MAP);
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				ChangeDescription description = recorder.endRecording();
				if (!(description.eContents().isEmpty()) && !sourceClass.isAbstract()) {
					propagateChanges(description, sourceClass, sourceClass.eResource());
				}
			}

		}
		return null;

	}

	/**
	 * If a Node is typed over the SuperClass and contains the pulled down Attribute, 
	 * retypes it to a SubClass chosen by User. 
	 */
	@SuppressWarnings({ "rawtypes" })
	@Override
	protected void applyToAffectedObjects(ENamedElement eNamedElement) {
		for (Resource res : gtResources) {
			EcoreUtil.resolveAll(res);
			TreeIterator<EObject> iter = res.getAllContents();
			EList<EObject> objectsToRemove = new BasicEList<EObject>();
			String ruleName = "";
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					ruleName = ((GraphTransformationRule) eObject).getName();
					propagation.propagatePushDownAttribute(attr, sourceClass, null, (GraphTransformationRule) eObject);
				}
				else if (eObject instanceof GTField	&& ((GTField) eObject).getKind().getName().equals("OCL_CONSTRAINT")) {
					String containerName = "OCL_CONSTRAINT";
					String expression = ((GTField) eObject).getContent();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, attrName, containerName);
					if(newExpression != null)
						((GTField) eObject).setContent(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if (eObject instanceof GTPath) {
					String containerName = "GTPath";
					String expression = ((GTPath) eObject).getExpression();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, attrName, containerName);
					if(newExpression != null)
						((GTPath) eObject).setExpression(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPrecondition){
					String containerName = "GTPrecondition";
					String expression = ((GTPrecondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, attrName, containerName);
					if(newExpression != null)
						((GTPrecondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPostcondition){
					String containerName = "GTPostcondition";
					String expression = ((GTPostcondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, attrName, containerName);
					if(newExpression != null)
						((GTPostcondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof Shape && objectsToRemove.contains(((Shape)eObject).getElement()))
					objectsToRemove.add(eObject);
			}
			propagation.cleanUpAndSave(objectsToRemove, res);
			
			URI resURI = res.getURI();
			String fragment = resURI.toString();
			String[] resRuleNameArray = fragment.split("/");
			String resRuleName = resRuleNameArray[resRuleNameArray.length-1];			
			int response = JOptionPane.showConfirmDialog(null,
					"Would you like to validate the GT rule " + resRuleName + "?", "Attention",
					JOptionPane.YES_NO_OPTION);
			if(response == 0){
				try{
					boolean result = callValidation(res);
					if(result == false){
						JOptionPane.showMessageDialog(null,
								"Errors exist in the rule " + resRuleName + "\nPlease check it manually.", "Validation failed!",
								JOptionPane.ERROR_MESSAGE);
					}
				} catch (Throwable e){
					e.printStackTrace();
				}
			}
		}
	}	
	/**
	 * If a Node is typed over the SuperClass and contains the pulled down Attribute, 
	 * retypes it to a SubClass chosen by User. 
	 * Not used at the moment.
	 */
	@SuppressWarnings({ "rawtypes", "unchecked" })
	protected void applyToAffectedObjects2(	ENamedElement eNamedElement) {
		for (Resource res : gtResources) {
			EcoreUtil.resolveAll(res);
			TreeIterator<EObject> iter = res.getAllContents();
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					List<GPElement> patternElements = ((GraphTransformationRule) eObject).getPattern().getElements();
					List<GTNegativeApplicationCondition> nacList = ((GraphTransformationRule) eObject).getNegativePattern();
					List<EClass> subClasses = helper.getAllSubClasses(sourceClass);
					List<String> subClassesNames = new ArrayList<>();
					for(EClass c : subClasses)
						subClassesNames.add(c.getName());
					for(GPElement e : patternElements){
						if(e instanceof GPUnboundNode && ((EClass) ((GPUnboundNode) e).getType()).getName().equals(sourceClass.getName())){
							for(GTAttribute a : ((GPUnboundNode) e).getAttributes())
								if(a.getExAttribute().eIsProxy()){
									String targetClassName = getSelectedFromDialog(subClassesNames, "Choose the new type for the node " + ((GPUnboundNode) e).getName() + " in the " + ((GraphTransformationRule) eObject).getName() + " GTRule");
									EClass targetClass = helper.getClass(targetClassName);
									((GPUnboundNode) e).setType(targetClass);
									for(EAttribute eAttr : ((EClass) ((GPUnboundNode) e).getType()).getEAttributes()){
										if(eAttr.getName().equals(attrName))
											a.setExAttribute(eAttr);
									}
								}
						}
						else if (e instanceof GTThisObject && ((EClass) ((GTThisObject) e).getType()).getName().equals(sourceClass.getName())){
							for(GTAttribute a : ((GTThisObject) e).getAttributes())
								if(a.getExAttribute().eIsProxy()){
									String targetClassName = getSelectedFromDialog(subClassesNames, "Choose the new type for the node " + ((GTThisObject) e).getName() + " in the " + ((GraphTransformationRule) eObject).getName() + " GTRule");
									EClass targetClass = helper.getClass(targetClassName);
									((GTThisObject) e).setType(targetClass);
									for(EAttribute eAttr : ((EClass) ((GTThisObject) e).getType()).getEAttributes()){
										if(eAttr.getName().equals(attrName))
											a.setExAttribute(eAttr);
									}
								}
						}
						else if (e instanceof GPBoundNode && ((EClass) ((GPBoundNode) e).getType()).getName().equals(sourceClass.getName())){
							for(GTAttribute a : ((GPBoundNode) e).getAttributes())
								if(a.getExAttribute().eIsProxy()){
									String targetClassName = getSelectedFromDialog(subClassesNames, "Choose the new type for the node " + ((GPBoundNode) e).getName() + " in the " + ((GraphTransformationRule) eObject).getName() + " GTRule");
									EClass targetClass = helper.getClass(targetClassName);
									((GPBoundNode) e).setType(targetClass);
									for(EAttribute eAttr : ((EClass) ((GPBoundNode) e).getType()).getEAttributes()){
										if(eAttr.getName().equals(attrName))
											a.setExAttribute(eAttr);
									}
								}
						}
					}
					for(GTNegativeApplicationCondition nac : nacList){
						List<NACElement> nacElements = nac.getElements();
						for(NACElement e : nacElements){
							if(e instanceof NACObject && ((EClass) ((NACObject) e).getType()).getName().equals(sourceClass.getName())){
								for(GTAttribute a : ((NACObject) e).getAttributes())
									if(a.getExAttribute().eIsProxy()){
										String targetClassName = getSelectedFromDialog(subClassesNames, "Choose the new type for the node " + ((NACObject) e).getName() + " in the " + ((GraphTransformationRule) eObject).getName() + " GTRule");
										EClass targetClass = helper.getClass(targetClassName);
										((NACObject) e).setType(targetClass);
										for(EAttribute eAttr : ((EClass) ((NACObject) e).getType()).getEAttributes()){
											if(eAttr.getName().equals(attrName))
												a.setExAttribute(eAttr);
										}										
									}
							}
							else if(e instanceof NACBoundNode){
								for(GTAttribute a : ((NACBoundNode) e).getAttributes())
									if(a.getExAttribute().eIsProxy())
										for(EAttribute eAttr : ((EClass) ((NACBoundNode) e).getGPUnboundNode().getType()).getEAttributes()){
											if(eAttr.getName().equals(attrName))
												a.setExAttribute(eAttr);
										}
							}
							else if (e instanceof GTThisObject && ((EClass) ((GTThisObject) e).getType()).getName().equals(sourceClass.getName())){
								for(GTAttribute a : ((GTThisObject) e).getAttributes())
									if(a.getExAttribute().eIsProxy()){
										String targetClassName = getSelectedFromDialog(subClassesNames, "Choose the new type for the node " + ((GTThisObject) e).getName() + " in the " + ((GraphTransformationRule) eObject).getName() + " GTRule");
										EClass targetClass = helper.getClass(targetClassName);
										((GTThisObject) e).setType(targetClass);
										for(EAttribute eAttr : ((EClass) ((GTThisObject) e).getType()).getEAttributes()){
											if(eAttr.getName().equals(attrName))
												a.setExAttribute(eAttr);
										}
									}
							}
						}
					}
				}
			}
			propagation.cleanUpAndSave(null, res);
		}
	}	
}