package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.regex.Pattern;

import javax.swing.JOptionPane;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EParameter;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.presentation.EcoreEditor;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.gmf.runtime.notation.Shape;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;

import de.ubt.ai1.modgraph.gt.failure.GTFailure;
import de.ubt.ai1.modgraph.gt.gtEngine.GPElement;
import de.ubt.ai1.modgraph.gt.gtEngine.GPUnboundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GTField;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPath;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPostcondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPrecondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTThisObject;
import de.ubt.ai1.modgraph.gt.gtEngine.GraphTransformationRule;

public class PushDownEOperationConnector extends RefactoringConnector {
	private IEditorPart activeEditor;
	private EOperation op;
	private String opName;
	private EClass sourceClass;

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		activeEditor = HandlerUtil.getActiveEditor(event);
		window = HandlerUtil.getActiveWorkbenchWindow(event);

		if (activeEditor instanceof EcoreEditor) {
			EcoreEditor ecoreEditor = (EcoreEditor) activeEditor;
			ecoreEditor.doSave(null);
			TreeSelection sel = (TreeSelection) ecoreEditor.getSelection();
			Object obj = sel.getFirstElement();

			if (obj instanceof EOperation) {
				op = (EOperation) obj;
				opName = op.getName();
				sourceClass = (EClass) op.eContainer();
				List<EParameter> params = op.getEParameters();
				EClassifier[] paramTypes = new EClassifier[params.size()];
				String name = op.getName();
				for (int i = 0; i < params.size(); i++) {
					paramTypes[i] = params.get(i).getEType();
				}
				initRefactoring(sourceClass.getEPackage().eContents());
				ChangeRecorder recorder = new ChangeRecorder(sourceClass);
				try {
					refactoring.pushDownEOperation(op.getName(),
							sourceClass.getName());
					for (EClass c : helper.getAllSubClasses(sourceClass)) {
						for (EOperation o : c.getEOperations()) {
							if (o.getName().equals(name)) {
								int j = 0;
								for (EParameter p : o.getEParameters()) {
									p.setEType(paramTypes[j]);
									j++;
								}
							}
						}
					}
				} catch (Throwable e) {
					JOptionPane.showMessageDialog(null, "Refactoring failed.",
							"ERROR!", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
					return null;
				}
				try {
					sourceClass.getEPackage().eResource().save(Collections.EMPTY_MAP);
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				ChangeDescription description = recorder.endRecording();
				if (!(description.eContents().isEmpty()) && !sourceClass.isAbstract()) {
					propagateChanges(description, sourceClass, sourceClass.eResource());
				}
			}

		}
		return null;

	}


	@SuppressWarnings("rawtypes")
	@Override
	protected void applyToAffectedObjects(ENamedElement eNamedElement) {
		for (Resource res : gtResources) {
			EcoreUtil.resolveAll(res);
			TreeIterator<EObject> iter = res.getAllContents();
			String ruleName = "";
			EList<EObject> objectsToRemove = new BasicEList<EObject>();
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					ruleName = ((GraphTransformationRule) eObject).getName();
					propagation.propagatePushDownOperation(op, sourceClass, null, (GraphTransformationRule) eObject);					
				}
				else if (eObject instanceof GTField	&& ((GTField) eObject).getKind().getName().equals("OCL_CONSTRAINT")) {
					String containerName = "OCL_CONSTRAINT";
					String expression = ((GTField) eObject).getContent();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTField) eObject).setContent(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if (eObject instanceof GTPath) {
					String containerName = "GTPath";
					String expression = ((GTPath) eObject).getExpression();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTPath) eObject).setExpression(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPrecondition){
					String containerName = "GTPrecondition";
					String expression = ((GTPrecondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTPrecondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPostcondition){
					String containerName = "GTPostcondition";
					String expression = ((GTPostcondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(ruleName, expression, eNamedElement.getName(), containerName);
					if(newExpression != null)
						((GTPostcondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof Shape && objectsToRemove.contains(((Shape)eObject).getElement()))
					objectsToRemove.add(eObject);
			}
			propagation.cleanUpAndSave(objectsToRemove, res);
			
			URI resURI = res.getURI();
			String fragment = resURI.toString();
			String[] resRuleNameArray = fragment.split("/");
			String resRuleName = resRuleNameArray[resRuleNameArray.length-1];			
			int response = JOptionPane.showConfirmDialog(null,
					"Would you like to validate the GT rule " + resRuleName + "?", "Attention",
					JOptionPane.YES_NO_OPTION);
			if(response == 0){
				try{
					boolean result = callValidation(res);
					if(result == false){
						JOptionPane.showMessageDialog(null,
								"Errors exist in the rule " + resRuleName + "\nPlease check it manually.", "Validation failed!",
								JOptionPane.ERROR_MESSAGE);
					}
				} catch (Throwable e){
					e.printStackTrace();
				}
			}
		}
	}	


	@SuppressWarnings({ "rawtypes", "unchecked" })
	protected void applyToAffectedObjects2(ENamedElement eNamedElement) {
		for (Resource res : gtResources) {
			EcoreUtil.resolveAll(res);
			TreeIterator<EObject> iter = res.getAllContents();
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					List<GPElement> patternElements = ((GraphTransformationRule) eObject).getPattern().getElements();
					List<EClass> subClasses = helper.getAllSubClasses(sourceClass);
					List<String> subClassesNames = new ArrayList<>();
					for(EClass c : subClasses)
						subClassesNames.add(c.getName());
					EObject implOp = ((GraphTransformationRule) eObject).getImplementedOperation();
					String implOpName = "";
					if(implOp.eIsProxy()){
						URI implOpURI = EcoreUtil.getURI(implOp);
						String fragment = implOpURI.fragment();
						implOpName = fragment.split("/")[fragment.split("/").length - 1];
						String targetClassName = getSelectedFromDialog(subClassesNames, "Choose the new CorrespondingClass  for the " + ((GraphTransformationRule) eObject).getName() + " GTRule");
						EClass targetClass = helper.getClass(targetClassName);
						((GraphTransformationRule) eObject).setCorrespondingClass(targetClass);
						for(EOperation op : targetClass.getEOperations())
							if(op.getName().equals(implOpName))
								((GraphTransformationRule) eObject).setImplementedOperation(op);
					} else {
						implOpName = ((EOperation) implOp).getName();
					}
					for(GPElement e : patternElements){
						if (e instanceof GPUnboundNode && ((EClass) ((GPUnboundNode) e).getType()).getName().equals(sourceClass.getName())){
							for(GTField f : ((GPUnboundNode) e).getFields()){
								if(((GTField)f).getKind().getName().equals("OPERATION_CALL")){
									String content = ((GTField) f).getContent();
									String operationName = content.split(Pattern.quote("("))[0];
									if(operationName.equals(opName)){
										String targetClassName = getSelectedFromDialog(subClassesNames, "Choose the new type for the node " + ((GPUnboundNode) e).getName() + " in the " + ((GraphTransformationRule) eObject).getName() + " GTRule");
										EClass targetClass = helper.getClass(targetClassName);
										((GPUnboundNode) e).setType(targetClass);										
									}									
								}									
							}
						}
						else if (e instanceof GTThisObject && ((EClass) ((GTThisObject) e).getType()).getName().equals(sourceClass.getName())){
							for(GTField f : ((GTThisObject) e).getFields()){
								if(((GTField)f).getKind().getName().equals("OPERATION_CALL")){
									String content = ((GTField) f).getContent();
									String operationName = content.split(Pattern.quote("("))[0];
									if(operationName.equals(opName)){
										String targetClassName = getSelectedFromDialog(subClassesNames, "Choose the new type for the node " + ((GTThisObject) e).getName() + " in the " + ((GraphTransformationRule) eObject).getName() + " GTRule");
										EClass targetClass = helper.getClass(targetClassName);
										((GTThisObject) e).setType(targetClass);										
									}									
								}									
							}
						}
					}					
				}
			}
			propagation.cleanUpAndSave(null, res);
//			vmh.initValidation(gtFiles.get(gtResources.indexOf(res)), window, new NullProgressMonitor());
		}
	}
}