package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.io.IOException;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JOptionPane;
import javax.swing.UIManager;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IHandler;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.FeatureChange;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchWindow;

import de.ubt.ai1.modGraph.example.eRefactoring.ERefactoringFactory;
import de.ubt.ai1.modGraph.example.eRefactoring.Helper;
import de.ubt.ai1.modGraph.example.eRefactoring.Propagation;
import de.ubt.ai1.modGraph.example.eRefactoring.Refactoring;
import de.ubt.ai1.modgraph.gt.gtEngine.GraphTransformationRule;
import de.ubt.ai1.modgraph.validation.handler.ValidationMenuHandler;

/**
 * 
 * @author Sabine Winetzhammer
 * 
 */
@SuppressWarnings({ "rawtypes", "unchecked" })
public abstract class RefactoringConnector extends AbstractHandler implements
		IHandler {
	protected IEditorPart activeEditor;
	protected IWorkbenchWindow window;
	protected Refactoring refactoring = ERefactoringFactory.eINSTANCE
			.createRefactoring();
	protected Helper helper = ERefactoringFactory.eINSTANCE.createHelper();
	protected Propagation propagation = ERefactoringFactory.eINSTANCE
			.createPropagation();
	ValidationMenuHandler vmh = new ValidationMenuHandler();
	List<Resource> gtResources = new LinkedList<>();
	List<IFile> gtFiles = new LinkedList<>();

	/** Implement this in subclass!!! **/
	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		return null;
	}

	public String getSelectedFromDialog(List<String> pos, String headding) {

		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {
			System.out.println("Error setting native look and feel: " + e);
		}
		String s = (String) javax.swing.JOptionPane.showInputDialog(null,
				headding + ":\n", "Choose", JOptionPane.PLAIN_MESSAGE, null,
				pos.toArray(), null);
		return s;
	}

	protected void initRefactoring(List<EObject> objects) {
		for (EObject o : objects) {
			if (o instanceof EClass) {
				EClass temp = (EClass) o;
				refactoring.getReferenceToEClass().add(temp);
				helper.getReferenceToEClassHelper().add(temp);
				List<EOperation> op = temp.getEOperations();
				List<EAttribute> at = temp.getEAttributes();
				List<EReference> re = temp.getEReferences();
				for (EOperation a : op) {
					refactoring.getReferenceToEOperation().add(a);
					helper.getReferenceToEOperationHelper().add(a);
					refactoring.getReferenceToEParameter().addAll(
							a.getEParameters());
					helper.getReferenceToEParameterHelper().addAll(
							a.getEParameters());
				}
				for (EAttribute a : at) {
					refactoring.getReferenceToEAttribute().add(a);
					helper.getReferenceToEAttributeHelper().add(a);
				}
				for (EReference a : re) {
					refactoring.getReferenceToEReference().add(a);
					helper.getReferenceToEReferenceHelper().add(a);
				}
			}
		}
		refactoring.setHelper(helper);
		propagation.setHelper(helper);
	}

	protected boolean propagateChanges(ChangeDescription description,
			ENamedElement eNamedElement, Resource res) {
		EList<FeatureChange> changes = description.getObjectChanges().get(
				eNamedElement);
		EList<EObject> addedObjects = description.getObjectsToAttach();
		EList<EObject> removedObjects = description.getObjectsToDetach();
		gtResources.clear();
		getGTResources(res);
		if (changes != null)
			for (FeatureChange change : changes) {
				System.out.println(change + " Name: " + change.getFeatureName()
						+ " Value: " + change.getDataValue());
			}
		if (addedObjects != null)
			for (EObject addedObj : addedObjects) {
				System.out.println(addedObj);
			}
		if (removedObjects != null)
			for (EObject removedObj : removedObjects) {
				System.out.println(removedObj);
			}
		applyToAffectedObjects(eNamedElement);
		return false;
	}

	protected void getGTResources(Resource res) {

		URI ecoreURI = res.getURI();
		IFile ecoreIFile = ResourcesPlugin.getWorkspace().getRoot()
				.getFile(new Path(ecoreURI.toPlatformString(true)));
		IProject project = ecoreIFile.getProject();
		ResourceSet rs = new ResourceSetImpl();

		recursiveGetGTResources(project, rs);
		EcoreUtil.resolveAll(rs);

	}

	private List<Resource> recursiveGetGTResources(IResource iresource,
			ResourceSet rs) {
		try {

			for (IResource ires : ((IContainer) iresource).members()) {
				// System.out.println(ires);
				if (ires instanceof IFolder && !(ires.getName().equals("bin"))
						&& !(ires.getName().equals("src")))
					// gtResources.addAll(
					recursiveGetGTResources(ires, rs);

				if (ires instanceof IFile && ires.getFileExtension() != null
						&& ires.getFileExtension().equals("gtEngine")) {
					URI gtURI = URI.createPlatformResourceURI(ires
							.getFullPath().toString(), true);
					Resource gtRes = rs.getResource(gtURI, true);
					gtResources.add(gtRes);
					gtFiles.add((IFile) ires);
					System.out.println(ires);
					System.out.println(gtRes.getContents());
				}
			}
			return gtResources;
		} catch (CoreException e) {

			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Depends on the refacoring performed.
	 * 
	 * @param eNamedElement
	 * @return
	 */
	protected abstract void applyToAffectedObjects(ENamedElement eNamedElement);

	protected void changeCorrespondingClass(EObject eObject,
			ENamedElement eNamedElement, FeatureChange change, Resource res) {

		GraphTransformationRule<EObject, EObject> rule = ((GraphTransformationRule) eObject);

		EClass changeClass = (EClass) change.getValue();

		if (rule.getCorrespondingClass().eIsProxy()
				|| ((ENamedElement) rule.getCorrespondingClass()).getName()
						.equals(changeClass.getName())) {

			URI corrClassURI = EcoreUtil.getURI(rule.getCorrespondingClass());

			String fragment = corrClassURI.fragment();
			String className = fragment.split("/")[fragment.split("/").length - 1];

			if (changeClass.getName().equals(className)) {

				// impl. op not known anymore...
				if (rule.getImplementedOperation().eIsProxy()) {
					URI implOpURI = EcoreUtil.getURI(rule
							.getImplementedOperation());

					String opfragment = implOpURI.fragment();
					String opName = opfragment.split("/")[fragment.split("/").length];

					for (EObject o : eNamedElement.eContents()) {
						if (o instanceof EOperation
								&& ((ENamedElement) o).getName().equals(opName)) {
							rule.setCorrespondingClass(eNamedElement);
							rule.setImplementedOperation(o);
							try {
								res.save(Collections.EMPTY_MAP);
							} catch (IOException e) {

								e.printStackTrace();
							}
							break;
						}
					}
				}
				try {
					res.save(Collections.EMPTY_MAP);
				} catch (IOException e) {

					e.printStackTrace();
				}
			}
		}
	}

	protected void changeImplementedOperation(EObject eObject,
			ENamedElement eNamedElement, FeatureChange change, Resource res) {

		GraphTransformationRule<EObject, EObject> rule = ((GraphTransformationRule) eObject);
		if (rule.getImplementedOperation().eIsProxy()
				|| ((ENamedElement) rule.getImplementedOperation()).getName()
						.equals(change.getDataValue())) {
			URI implOpURI = EcoreUtil.getURI(rule.getImplementedOperation());

			String fragment = implOpURI.fragment();
			String opName = fragment.split("/")[fragment.split("/").length - 1];

			if (change.getDataValue().equals(opName)) {
				rule.setImplementedOperation(eNamedElement);
			}

		}
		try {
			res.save(Collections.EMPTY_MAP);
		} catch (IOException e) {

			e.printStackTrace();
		}
	}

	protected boolean callValidation(Resource res) throws ExecutionException {
		return vmh.initValidation(gtFiles.get(gtResources.indexOf(res)),
				window, new NullProgressMonitor());
	}

	// /**
	// * Move corresponding shape in the diagram.
	// * @param res
	// * @param movedObject
	// * @param newContainer
	// * @param type
	// */
	// protected void moveShape(Resource res, EObject movedObject, EObject
	// newContainer, String type) {
	// TreeIterator<EObject> iter2 = res.getAllContents();
	// Shape shape = null;
	// Node targetNode = null;
	// String[] constraintOperatorsArray = {"==", "!=", "<=", ">=", "<", ">"};
	// List<String> constraintOperators =
	// Arrays.asList(constraintOperatorsArray);
	// while(iter2.hasNext()){
	// EObject next = iter2.next();
	// if(next instanceof Shape &&
	// ((Shape)next).getElement().equals(movedObject))
	// shape = (Shape) next;
	// if(next instanceof Node && ((Node)next).getType().equals(type) &&
	// ((Node)next).getElement().equals(newContainer))
	// targetNode = (Node) next;
	// }
	// if(shape != null && targetNode != null){
	// for(Object o : targetNode.getPersistedChildren()){
	// if(o instanceof BasicCompartment && shape.getElement() instanceof
	// GTAttribute
	// && constraintOperators.contains(((GTAttribute)
	// shape.getElement()).getOperator().getLiteral())){
	// ((BasicCompartment)o).getPersistedChildren().add(shape);
	// break;
	// }
	// else if(o instanceof BasicCompartment)
	// ((BasicCompartment)o).getPersistedChildren().add(shape);
	// }
	// }
	// }
}
