package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.swing.JOptionPane;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IHandler;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EParameter;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.presentation.EcoreEditor;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.gmf.runtime.notation.Edge;
import org.eclipse.gmf.runtime.notation.Node;
import org.eclipse.gmf.runtime.notation.Shape;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;

import de.ubt.ai1.modgraph.gt.gtEngine.GPParameter;
import de.ubt.ai1.modgraph.gt.gtEngine.GTField;
import de.ubt.ai1.modgraph.gt.gtEngine.GTLink;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPath;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPostcondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPrecondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GraphTransformationRule;
import de.ubt.ai1.modgraph.gt.gtEngine.NACParameter;

public class RemoveEParameterConnector extends RefactoringConnector implements
		IHandler {
	private IEditorPart activeEditor;
	private String paramName = "";

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		activeEditor = HandlerUtil.getActiveEditor(event);
		window = HandlerUtil.getActiveWorkbenchWindow(event);

		if (activeEditor instanceof EcoreEditor) {
			EcoreEditor ecoreEditor = (EcoreEditor) activeEditor;
			ecoreEditor.doSave(null);
			TreeSelection sel = (TreeSelection) ecoreEditor.getSelection();
			Object obj = sel.getFirstElement();

			if (obj instanceof EOperation) {
				EOperation op = (EOperation) obj;
				List<String> paramNames = new LinkedList<String>();
				EList<EParameter> params = op.getEParameters();
				for (EParameter p : params)
					paramNames.add(p.getName());				

				paramName = getSelectedFromDialog(paramNames,
						"Choose the parameter to be removed.");
				
				ChangeRecorder recorder = new ChangeRecorder();
				Map<EObject, URI> eObjectToProxyURIMap = new HashMap<EObject, URI>();
				recorder.setEObjectToProxyURIMap(eObjectToProxyURIMap);
				recorder.beginRecording(Collections.singleton(op));

				initRefactoring(op.getEContainingClass().getEPackage()
						.eContents());

				try {
					refactoring.removeEParameter(op.getName(), paramName);
				} catch (Throwable e) {
					JOptionPane.showMessageDialog(null, "Refactoring failed.",
							"ERROR!", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
					return null;
				}
				try {
					op.getEContainingClass().getEPackage().eResource()
					.save(Collections.EMPTY_MAP);
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				
				ChangeDescription description = recorder.endRecording();
				if (!(description.eContents().isEmpty())) {
					propagateChanges(description, op, op.eResource());
//					propagateChanges2(description, op);
				}				
			}

		}
		return null;

	}
	
	/** 
	 * Checks that the rule is the right one,
	 * then removes the parameter and its reference from the rule.
	 * Updates the OCL Expressions if needed.
	 * Changes OPERATION_CALLs according to the corresponding EOperation's parameter list.
	 * Removes corresponding nodes and edges from the GTRule's diagram.
	 */
	@SuppressWarnings({ "rawtypes" })
	@Override
	protected void applyToAffectedObjects(ENamedElement eNamedElement) {
		EList<Object> temp = new BasicEList<Object>();
		for (Resource res : gtResources) 
			temp.add((Object) res);
		for (Resource res : gtResources) {
			TreeIterator<EObject> iter = res.getAllContents();
			EList<EObject> objectsToRemove = new BasicEList<EObject>();
			GraphTransformationRule rule = null;
			while (iter.hasNext()){
				EObject eObject = iter.next();
				if(eObject instanceof  GraphTransformationRule<?, ?>){
					rule = (GraphTransformationRule) eObject;
					propagation.propagateRemoveParameterInit(paramName, eNamedElement.getName(), (GraphTransformationRule) eObject);
				}
				else if (eObject instanceof GTField	&& ((GTField) eObject).getKind().getName().equals("OCL_CONSTRAINT")) {
					String containerName = "OCL_CONSTRAINT";
					String expression = ((GTField) eObject).getContent();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(rule.getName(), expression, paramName, containerName);
					if(newExpression != null)
						((GTField) eObject).setContent(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if (eObject instanceof GTPath) {
					String containerName = "GTPath";
					String expression = ((GTPath) eObject).getExpression();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(rule.getName(), expression, paramName, containerName);
					if(newExpression != null)
						((GTPath) eObject).setExpression(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPrecondition){
					String containerName = "GTPrecondition";
					String expression = ((GTPrecondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(rule.getName(), expression, paramName, containerName);
					if(newExpression != null)
						((GTPrecondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof GTPostcondition){
					String containerName = "GTPostcondition";
					String expression = ((GTPostcondition) eObject).getCondition();
					String newExpression = propagation.updateOCLExpressionWithUserInteraction(rule.getName(), expression, paramName, containerName);
					if(newExpression != null)
						((GTPostcondition) eObject).setCondition(newExpression);
					if(newExpression != null && newExpression.equals(""))
						objectsToRemove.add(eObject);
				}
				else if(eObject instanceof Shape && objectsToRemove.contains(((Shape)eObject).getElement()))
					objectsToRemove.add(eObject);
			}
			TreeIterator<EObject> iter2 = res.getAllContents();
			while (iter2.hasNext()){
				EObject eObject = iter2.next();
				if(eObject instanceof Node){
					String type = ((Node)eObject).getType();
					if(type.equals("2009") || type.equals("2010") || type.equals("2016")){
						EObject element = ((Node) eObject).getElement();
						if(!(element instanceof GPParameter || element instanceof NACParameter ))
							objectsToRemove.add(eObject);						
					}
				}
				else if(eObject instanceof Edge){
					EObject element = ((Edge) eObject).getElement();
					if(!(element instanceof GTLink))
						objectsToRemove.add(eObject);
				}
			}		
			propagation.updateOperationCalls(temp);
			propagation.cleanUpAndSave(objectsToRemove, res);
			
			URI resURI = res.getURI();
			String fragment = resURI.toString();
			String[] resRuleNameArray = fragment.split("/");
			String resRuleName = resRuleNameArray[resRuleNameArray.length-1];			
			int response = JOptionPane.showConfirmDialog(null,
					"Would you like to validate the GT rule " + resRuleName + "?", "Attention",
					JOptionPane.YES_NO_OPTION);
			if(response == 0){
				try{
					boolean result = callValidation(res);
					if(result == false){
						JOptionPane.showMessageDialog(null,
								"Errors exist in the rule " + resRuleName + "\nPlease check it manually.", "Validation failed!",
								JOptionPane.ERROR_MESSAGE);
					}
				} catch (Throwable e){
					e.printStackTrace();
				}
			}
		}
	}

//	protected boolean propagateChanges2(ChangeDescription description,
//			ENamedElement eNamedElement) {
//		EList<FeatureChange> changes = description.getObjectChanges().get(
//				eNamedElement);
//		EList<EObject> addedObjects = description.getObjectsToAttach();
//		EList<EObject> removedObjects = description.getObjectsToDetach();
//		getGTResources(eNamedElement);
//		if (changes != null)
//			for (FeatureChange change : changes) {
//				System.out.println(change + " Name: " + change.getFeatureName()
//						+ " Value: " + change.getDataValue());			
//				applyToAffectedObjects2(change, eNamedElement);
//			}
//		if (addedObjects != null)
//			for (EObject addedObj : addedObjects) {
//				System.out.println(addedObj);
//			
//				
//			}
//		if (removedObjects != null)
//			for (EObject removedObj : removedObjects) {
//				System.out.println(removedObj);
//							}
//		return false;
//	}
//	
//	/** 
//	 * Checks that the rule is the right one,
//	 * then removes the parameter and its reference from the rule.
//	 * Changes OPERATION_CALLs according to the corresponding EOperation's parameter list.
//	 * Removes corresponding nodes and edges from the GTRule's diagram.
//	 */
//	@SuppressWarnings({ "rawtypes", "unchecked" })
//	protected List<EObject> applyToAffectedObjects2(FeatureChange change,
//			ENamedElement eNamedElement) {
//
//		for (Resource res : gtResources) {
//			EcoreUtil.resolveAll(res);
//			TreeIterator<EObject> iter = res.getAllContents();
//			List<EObject> objectsToRemove = new ArrayList<EObject>();
//			while (iter.hasNext()){
//				EObject eObject = iter.next();
//				if(eObject instanceof GPParameter){
//					if(((GPParameter) eObject).getName().equals(paramName)){
//						GraphTransformationRule<EObject, EObject> rule = (GraphTransformationRule<EObject, EObject>) ((GPParameter<?>) eObject).eContainer().eContainer();
//						EOperation ruleOperation = (EOperation) rule.getImplementedOperation();
//						if(ruleOperation.getName().equals(eNamedElement.getName())){
//							objectsToRemove.add(eObject);
//							objectsToRemove.addAll(((GPParameter) eObject).getIncomingEdges());
//							objectsToRemove.addAll(((GPParameter) eObject).getOutgoingEdges());
//						}						
//					}
//				}
//				else if(eObject instanceof GTField && ((GTField) eObject).getKind().getName().equals("OPERATION_CALL")){
//					String content = ((GTField) eObject).getContent();
//					String operationName = content.split(Pattern.quote("("))[0];
//					if(operationName.equals(eNamedElement.getName())){
//						String newParamNames = "";
//						for(EParameter p : ((EOperation)eNamedElement).getEParameters()){
//							if (p != null && newParamNames.equals("")) {
//								newParamNames = p.getName();
//							} else if (p != null && !newParamNames.equals("")) {
//								newParamNames = newParamNames + "," + p.getName();
//							}
//						}
//						String newName = operationName + "(" + newParamNames + ")";
//						((GTField) eObject).setContent(newName);
//					}
//				}
//				else if(eObject instanceof Node){
//					if(objectsToRemove.contains(((Node) eObject).getElement()))
//						objectsToRemove.add(eObject);
//				}
//				else if(eObject instanceof Edge){
//					if(objectsToRemove.contains(((Edge) eObject).getElement()))
//						objectsToRemove.add(eObject);
//				}
//			}		
//			cleanUpAndSave(objectsToRemove, res);
//		}
//		return null;
//	}
}
