package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JOptionPane;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EParameter;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.presentation.EcoreEditor;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.ui.handlers.HandlerUtil;

import de.ubt.ai1.modgraph.gt.gtEngine.GTAttribute;
import de.ubt.ai1.modgraph.gt.gtEngine.GTField;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPostcondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTPrecondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GraphTransformationRule;

public class RenameENamedElementConnector extends RefactoringConnector {

	private String oldName = null;
	private String newName = null;

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {
		activeEditor = HandlerUtil.getActiveEditor(event);
		window = HandlerUtil.getActiveWorkbenchWindow(event);

		if (activeEditor instanceof EcoreEditor) {
			EcoreEditor ecoreEditor = (EcoreEditor) activeEditor;
			ecoreEditor.doSave(null);

			TreeSelection sel = (TreeSelection) ecoreEditor.getSelection();
			Object obj = sel.getFirstElement();

			if (obj instanceof ENamedElement) {
				ENamedElement namedElement = (ENamedElement) obj;
				refactoring.getReferenceToENamedElement().add(namedElement);

				// SW
				// ChangeRecorder recorder = new ChangeRecorder(namedElement);
				ChangeRecorder recorder = new ChangeRecorder();
				Map<EObject, URI> eObjectToProxyURIMap = new HashMap<EObject, URI>();
				recorder.setEObjectToProxyURIMap(eObjectToProxyURIMap);
				recorder.beginRecording(Collections.singleton(namedElement));

				oldName = namedElement.getName();
				newName = JOptionPane
						.showInputDialog("Enter the new name for the selected class.");
				if (newName.equals("")) {
					JOptionPane.showMessageDialog(null,
							"Selected name is not valid.", "ERROR!",
							JOptionPane.ERROR_MESSAGE);
					return null;
				}

				initRefactoring(namedElement.eResource().getContents());
				try {
					refactoring.renameENamedElement(namedElement,// .getName(),
							newName);
				} catch (Throwable e) {
					JOptionPane.showMessageDialog(null,
							"A class with the same name already exists.",
							"ERROR", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
					return null;
				}

				try {
					namedElement.eResource().save(Collections.EMPTY_MAP);
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				// SW
				ChangeDescription description = recorder.endRecording();
				if (!(description.eContents().isEmpty())) {
					propagateChanges(description, namedElement, namedElement.eResource());
				}
			}
		}
		return null;

	}

	@SuppressWarnings({ "rawtypes" })
	protected void applyToAffectedObjects(ENamedElement eNamedElement) {
		EcoreUtil.resolveAll(eNamedElement);
		EList<Object> temp = new BasicEList<Object>();		
		for (Resource res : gtResources) {
			temp.add((Object) res);
			TreeIterator<EObject> iter = res.getAllContents();
			GraphTransformationRule rule = null;
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					rule = (GraphTransformationRule) eObject;
					propagation.propagateRenameElement(eNamedElement, oldName,
							(GraphTransformationRule) eObject);
				}
				// change the value of gtAttributes if necessary
				else if (eObject instanceof GTAttribute){
					propagation.updateAttributeValue((GTAttribute) eObject, eNamedElement, oldName);
				}
				// changes the content of OPERATION_CALLs after renaming an eOperation
				else if(eObject instanceof GTField && ((GTField)eObject).getKind().getName().equals("OPERATION_CALL")){
					propagation.updateOperationCallWithNewName(eNamedElement, eObject, oldName);
				}
				// changes the OCL expressions of Preconditions
				else if(eObject instanceof GTPrecondition){
					if(!(eNamedElement instanceof EParameter) 
							|| (eNamedElement instanceof EParameter 
									&& propagation.compareOperations((EOperation)rule.getImplementedOperation(), (EOperation)eNamedElement.eContainer()))){
						String condition = ((GTPrecondition) eObject).getCondition();
						String newCondition = propagation.updateOCLExpressionWithNewName(condition, newName, oldName);
						if(newCondition != null)
							((GTPrecondition) eObject).setCondition(newCondition);	
					}
				}
				// changes the OCL expressions of Postconditions
				else if(eObject instanceof GTPostcondition){
					if(!(eNamedElement instanceof EParameter) 
							|| (eNamedElement instanceof EParameter 
									&& propagation.compareOperations((EOperation)rule.getImplementedOperation(), (EOperation)eNamedElement.eContainer()))){
						String condition = ((GTPostcondition) eObject).getCondition();
						String newCondition = propagation.updateOCLExpressionWithNewName(condition, newName, oldName);
						if(newCondition != null)
							((GTPostcondition) eObject).setCondition(newCondition);					
					}
				}
			}
			// updates operations calls just in case the parameters have changed
			propagation.updateOperationCalls(temp);
			// updates OCL Expressions in GTNodes and GTPaths
			if(!(eNamedElement instanceof EParameter) 
					|| (eNamedElement instanceof EParameter 
							&& propagation.compareOperations((EOperation)rule.getImplementedOperation(), (EOperation)eNamedElement.eContainer()))){
				propagation.updateOCLExpressions(temp, newName, oldName);
			}
			propagation.cleanUpAndSave(null, res);
			
			URI resURI = res.getURI();
			String fragment = resURI.toString();
			String[] resRuleNameArray = fragment.split("/");
			String resRuleName = resRuleNameArray[resRuleNameArray.length-1];			
			int response = JOptionPane.showConfirmDialog(null,
					"Would you like to validate the GT rule " + resRuleName + "?", "Attention",
					JOptionPane.YES_NO_OPTION);
			if(response == 0){
				try{
					boolean result = callValidation(res);
					if(result == false){
						JOptionPane.showMessageDialog(null,
								"Errors exist in the rule " + resRuleName + "\nPlease check it manually.", "Validation failed!",
								JOptionPane.ERROR_MESSAGE);
					}
				} catch (Throwable e){
					e.printStackTrace();
				}
			}
		}
	}	
}