package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.channels.FileChannel;
import java.util.Collections;

import javax.swing.JOptionPane;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IHandler;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.FeatureChange;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.presentation.EcoreEditor;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;

import de.ubt.ai1.modGraph.example.eRefactoring.ERefactoringFactory;
import de.ubt.ai1.modgraph.gt.gtEngine.GPBoundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GPElement;
import de.ubt.ai1.modgraph.gt.gtEngine.GPElementWithStatus;
import de.ubt.ai1.modgraph.gt.gtEngine.GPLink;
import de.ubt.ai1.modgraph.gt.gtEngine.GPUnboundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GTNegativeApplicationCondition;
import de.ubt.ai1.modgraph.gt.gtEngine.GTNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GTStatus;
import de.ubt.ai1.modgraph.gt.gtEngine.GTThisObject;
import de.ubt.ai1.modgraph.gt.gtEngine.GraphTransformationRule;
import de.ubt.ai1.modgraph.gt.gtEngine.GtEngineFactory;
import de.ubt.ai1.modgraph.gt.gtEngine.NACBoundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.NACElement;
import de.ubt.ai1.modgraph.gt.gtEngine.NACLink;
import de.ubt.ai1.modgraph.gt.gtEngine.NACObject;
import de.ubt.ai1.modgraph.gt.gtEngine.NACParameter;

public class ReplaceInheritanceWithDelegationConnector extends RefactoringConnector
		implements IHandler {
	private IEditorPart activeEditor;
	private EClass subClass;
	private EClass superClass;
	private EReference reference;
	private int i = 0;

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		activeEditor = HandlerUtil.getActiveEditor(event);
		window = HandlerUtil.getActiveWorkbenchWindow(event);

		if (activeEditor instanceof EcoreEditor) {
			EcoreEditor ecoreEditor = (EcoreEditor) activeEditor;
			ecoreEditor.doSave(null);
			TreeSelection sel = (TreeSelection) ecoreEditor.getSelection();
			Object obj = sel.getFirstElement();

			if (obj instanceof EClass) {
				subClass = (EClass) obj;
				superClass = subClass.getESuperTypes().get(0);

				ChangeRecorder recorder = new ChangeRecorder(subClass);
				initRefactoring(superClass.getEPackage().eContents());

				try {
					refactoring.replaceInheritanceWithDelegation(subClass.getName(), superClass.getName());				
				} catch (Throwable e) {
					JOptionPane.showMessageDialog(null, "Refactoring failed.","ERROR!", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
					return null;
				}
				try {
					subClass.getEPackage().eResource()
					.save(Collections.EMPTY_MAP);
				} catch (IOException e) {
					e.printStackTrace();
				}
				for(EReference r : subClass.getEReferences())
					if(r.getEType().getName().equals(superClass.getName()))
						reference = r;	

				ChangeDescription description = recorder.endRecording();
				if (!(description.eContents().isEmpty())) {
					propagateChanges(description, subClass, subClass.eResource());
				}
				
				
//				test();
			}

		}
		return null;

	}

	@SuppressWarnings({ "rawtypes" })
	@Override
	protected void applyToAffectedObjects(ENamedElement eNamedElement) {
		for (Resource res : gtResources) {
			TreeIterator<EObject> iter = res.getAllContents();
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					try {
						propagation.propagateRemoveMidMan_ReplInheritWDelegInit(reference, subClass, superClass, (GraphTransformationRule) eObject);
					} catch (Throwable e1) {
						e1.printStackTrace();
						break;
					}
				} 
			}
			
			URI resURI = res.getURI();
			String fragment = resURI.toString();
			String[] resRuleNameArray = fragment.split("/");
			String resRuleName = resRuleNameArray[resRuleNameArray.length-1];			
			int response = JOptionPane.showConfirmDialog(null,
					"Would you like to validate the GT rule " + resRuleName + "?", "Attention",
					JOptionPane.YES_NO_OPTION);
			if(response == 0){
				try{
					boolean result = callValidation(res);
					if(result == false){
						JOptionPane.showMessageDialog(null,
								"Errors exist in the rule " + resRuleName + "\nPlease check it manually.", "Validation failed!",
								JOptionPane.ERROR_MESSAGE);
					}
				} catch (Throwable e){
					e.printStackTrace();
				}
			}
		}
	}
	
//	
	@SuppressWarnings("unused")
	private void test() throws ExecutionException{
		File source = new File("C:\\Users\\Duemmel\\Desktop\\Projekt\\TimeTestModels\\BTReplInheritDeleg\\BugModel.ecore");
		File dest = new File("C:\\Users\\Duemmel\\runtime-EclipseApplication2\\Bugtracker\\model\\BugModel.ecore");
		FileChannel inputChannel = null;
		FileChannel outputChannel = null;
		try {
			inputChannel = new FileInputStream(source).getChannel();
			outputChannel = new FileOutputStream(dest).getChannel();
			outputChannel.transferFrom(inputChannel, 0, inputChannel.size());
		} catch (Throwable e) {	e.printStackTrace();} 
		finally {
			try {
				inputChannel.close();
				outputChannel.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		source = new File("C:\\Users\\Duemmel\\Desktop\\Projekt\\TimeTestModels\\BTReplInheritDeleg\\gt\\Ticket_assignTo.gtEngine");
		dest = new File("C:\\Users\\Duemmel\\runtime-EclipseApplication2\\Bugtracker\\model\\gt\\Ticket_assignTo.gtEngine");
		inputChannel = null;
		outputChannel = null;
		try {
			inputChannel = new FileInputStream(source).getChannel();
			outputChannel = new FileOutputStream(dest).getChannel();
			outputChannel.transferFrom(inputChannel, 0, inputChannel.size());
		} catch (Throwable e) {	e.printStackTrace();} 
		finally {
			try {
				inputChannel.close();
				outputChannel.close();
			} catch (IOException e) {e.printStackTrace();}
		}

		
		for(EReference r : subClass.getEReferences()){
			if(r.getEType().getName().equals(superClass.getName())){
				subClass.getEStructuralFeatures().remove(r);
			}
		}
		subClass.getESuperTypes().add(superClass);
		
		try {
			subClass.eResource().save(Collections.EMPTY_MAP);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		execute2();
	}
	
	private Object execute2() throws ExecutionException {
		

		refactoring = ERefactoringFactory.eINSTANCE.createRefactoring();
		helper = ERefactoringFactory.eINSTANCE.createHelper();
		propagation = ERefactoringFactory.eINSTANCE.createPropagation();
		initRefactoring(superClass.getEPackage().eContents());

		ChangeRecorder recorder = new ChangeRecorder(subClass);
		long preRefactoringNano = System.nanoTime();
		try {
			refactoring.replaceInheritanceWithDelegation(subClass.getName(), superClass.getName());		
		} catch (Throwable e) {
			JOptionPane.showMessageDialog(null, "Refactoring failed.","ERROR!", JOptionPane.ERROR_MESSAGE);
			e.printStackTrace();
			return null;
		}
//
//		EReference newRef = EcoreFactory.eINSTANCE.createEReference();
//		newRef.setEType(superClass);
//		newRef.setUpperBound(1);
//		newRef.setLowerBound(1);
//		newRef.setName(subClass.getName() + "to" + superClass.getName());
//		subClass.getEStructuralFeatures().add(newRef);
//		subClass.getESuperTypes().remove(superClass);
//		
		try {
			subClass.getEPackage().eResource().save(Collections.EMPTY_MAP);
		} catch (IOException e) {
			e.printStackTrace();
		}

		for(EReference r : subClass.getEReferences())
			if(r.getEType().getName().equals(superClass.getName()))
				reference = r;	

		long prePropagationNano = System.nanoTime();
		ChangeDescription description = recorder.endRecording();
		if (!(description.eContents().isEmpty())) {
			propagateChanges(description, subClass, subClass.eResource());
//			propagateChanges2(description, subClass);
		}
		long finalNano = System.nanoTime();
		
		try(PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter("C:/ReplaceInheritDeleg_Refactoring.txt", true)))) {
			out.println((prePropagationNano - preRefactoringNano));
		}catch (IOException e) {}			
		try(PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter("C:/ReplaceInheritDeleg_Propagation.txt", true)))) {
			out.println((finalNano - prePropagationNano));
		}catch (IOException e) {}
		

		for(EReference r : subClass.getEReferences()){
			if(r.getEType().getName().equals(superClass.getName())){
				subClass.getEStructuralFeatures().remove(r);
			}
		}		
		subClass.getESuperTypes().add(superClass);
		try {
			subClass.eResource().save(Collections.EMPTY_MAP);
		} catch (IOException e) {
			e.printStackTrace();
		}		
		
		File source = new File("C:\\Users\\Duemmel\\Desktop\\Projekt\\TimeTestModels\\BTReplInheritDeleg\\BugModel.ecore");
		File dest = new File("C:\\Users\\Duemmel\\runtime-EclipseApplication2\\Bugtracker\\model\\BugModel.ecore");
		FileChannel inputChannel = null;
		FileChannel outputChannel = null;
		try {
			inputChannel = new FileInputStream(source).getChannel();
			outputChannel = new FileOutputStream(dest).getChannel();
			outputChannel.transferFrom(inputChannel, 0, inputChannel.size());
		} catch (Throwable e) {	e.printStackTrace();} 
		finally {
			try {
				inputChannel.close();
				outputChannel.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		source = new File("C:\\Users\\Duemmel\\Desktop\\Projekt\\TimeTestModels\\BTReplInheritDeleg\\gt\\Ticket_assignTo.gtEngine");
		dest = new File("C:\\Users\\Duemmel\\runtime-EclipseApplication2\\Bugtracker\\model\\gt\\Ticket_assignTo.gtEngine");
		inputChannel = null;
		outputChannel = null;
		try {
			inputChannel = new FileInputStream(source).getChannel();
			outputChannel = new FileOutputStream(dest).getChannel();
			outputChannel.transferFrom(inputChannel, 0, inputChannel.size());
		} catch (Throwable e) {	e.printStackTrace();} 
		finally {
			try {
				inputChannel.close();
				outputChannel.close();
			} catch (IOException e) {e.printStackTrace();}
		}
		System.out.println(i);
		i++;		
		while(i<15){
			execute2();
		}
		
		return null;
	}
	
	
	//////////////////////////////////////////////////////////////////
	//	Complete Java Implementation for the sake of testing below	//
	//////////////////////////////////////////////////////////////////

	@SuppressWarnings("rawtypes")
	private void cleanUpAndSave2(EList<EObject> objectsToRemove, Resource res){
		TreeIterator<EObject> iter = res.getAllContents();
		EList<EObject> nacsToRemove = new BasicEList<EObject>();
		while (iter.hasNext()){
			EObject eObject = iter.next();
			if(eObject instanceof  GraphTransformationRule<?, ?>){
				for(Object nac : ((GraphTransformationRule)eObject).getNegativePattern()){
					if(((GTNegativeApplicationCondition)nac).getElements().isEmpty())
						{nacsToRemove.add((EObject) nac);}
				}
			}
		}
		if(objectsToRemove != null && !objectsToRemove.isEmpty()){
			for(EObject o : objectsToRemove){
				EcoreUtil.remove(o);
			}
		}
		if(!nacsToRemove.isEmpty()){
			for(EObject o : nacsToRemove){
				EcoreUtil.remove(o);
			}
		}
		try {
			res.save(java.util.Collections.EMPTY_MAP);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	@SuppressWarnings({ "rawtypes", "unchecked" })
	private void applyToAffectedObjects2(ENamedElement eNamedElement) {
		for (Resource res : gtResources) {
			TreeIterator<EObject> iter = res.getAllContents();
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					GraphTransformationRule rule = (GraphTransformationRule) eObject;
					EcoreUtil.resolveAll(eObject);
					EList<GPElement> patternElements = rule.getPattern().getElements();
					GTNode sourceNodePattern = null;
					GTNode targetNodePattern = null;
					String sourceClassName = subClass.getName();
					String targetClassName = superClass.getName();
					for(GPElement gpe : patternElements){
						if(gpe instanceof GTThisObject && ((EClass) ((GTThisObject)gpe).getType()).getName().equals(sourceClassName))
							sourceNodePattern = (GTNode) gpe;
						else if(gpe instanceof GPUnboundNode && ((EClass) ((GPUnboundNode)gpe).getType()).getName().equals(sourceClassName))
							sourceNodePattern = (GTNode) gpe;
						else if(gpe instanceof GPBoundNode && ((EClass) ((GPBoundNode)gpe).getType()).getName().equals(sourceClassName))
							sourceNodePattern = (GTNode) gpe;
						else if(gpe instanceof GTThisObject && ((EClass) ((GTThisObject)gpe).getType()).getName().equals(targetClassName))
							targetNodePattern = (GTNode) gpe;
						else if(gpe instanceof GPUnboundNode && ((EClass) ((GPUnboundNode)gpe).getType()).getName().equals(targetClassName))
							targetNodePattern = (GTNode) gpe;
						else if(gpe instanceof GPBoundNode && ((EClass) ((GPBoundNode)gpe).getType()).getName().equals(targetClassName))
							targetNodePattern = (GTNode) gpe;
					}
					if(sourceNodePattern != null && targetNodePattern != null){	
						GPLink newLink = GtEngineFactory.eINSTANCE.createGPLink();
						newLink.setExReference(reference);
						newLink.setSourceNode(sourceNodePattern);
						newLink.setTargetNode(targetNodePattern);
						if(targetNodePattern instanceof GPElementWithStatus){
							GTStatus status = ((GPElementWithStatus) targetNodePattern).getStatus();
							if(status.getName().equals("CREATE"))
								newLink.setStatus(GTStatus.CREATE);
							if(status.getName().equals("DELETE"))
								newLink.setStatus(GTStatus.DELETE);
						}
						patternElements.add(newLink);				
					}
					for(Object nac : rule.getNegativePattern()){
						GTNegativeApplicationCondition temp = (GTNegativeApplicationCondition) nac;
						EList<NACElement> nacElements = temp.getElements();
						GTNode sourceNodeNAC = null;
						GTNode targetNodeNAC = null;
						for(NACElement gpe : nacElements){
							if(gpe instanceof GTThisObject && ((EClass) ((GTThisObject)gpe).getType()).getName().equals(sourceClassName))
								sourceNodeNAC = (GTNode) gpe;
							else if(gpe instanceof NACObject && ((EClass) ((NACObject)gpe).getType()).getName().equals(sourceClassName))
								sourceNodeNAC = (GTNode) gpe;
							else if(gpe instanceof NACParameter && ((EClass) ((NACParameter)gpe).getType()).getName().equals(sourceClassName))
								sourceNodeNAC = (GTNode) gpe;
							else if(gpe instanceof NACBoundNode && ((EClass) ((NACBoundNode)gpe).getGPUnboundNode().getType()).getName().equals(sourceClassName))
								sourceNodeNAC = (GTNode) gpe;
							else if(gpe instanceof GTThisObject && ((EClass) ((GTThisObject)gpe).getType()).getName().equals(targetClassName))
								targetNodeNAC = (GTNode) gpe;
							else if(gpe instanceof NACObject && ((EClass) ((NACObject)gpe).getType()).getName().equals(targetClassName))
								targetNodeNAC = (GTNode) gpe;
							else if(gpe instanceof NACParameter && ((EClass) ((NACParameter)gpe).getType()).getName().equals(targetClassName))
								targetNodeNAC = (GTNode) gpe;
							else if(gpe instanceof NACBoundNode && ((EClass) ((NACBoundNode)gpe).getGPUnboundNode().getType()).getName().equals(targetClassName))
								targetNodeNAC = (GTNode) gpe;
						}
						if(sourceNodeNAC != null && targetNodeNAC != null){
							NACLink newLink = GtEngineFactory.eINSTANCE.createNACLink();
							newLink.setExReference(reference);
							newLink.setSourceNode(sourceNodeNAC);
							newLink.setTargetNode(targetNodeNAC);
							nacElements.add(newLink);
						}
					}
				} 
			}
			cleanUpAndSave2(null, res);
//			try{
//				boolean result = callValidation(res);
//				if(result == false){
//					JOptionPane.showMessageDialog(null,
//							"Errors exist in the rule " + ruleName + "\nPlease check this rule.", "Validation failed!",
//							JOptionPane.ERROR_MESSAGE);
//				}
//			} catch (Throwable e){
//				e.printStackTrace();
//			}
		}
	}
	
	@SuppressWarnings("unused")
	private boolean propagateChanges2(ChangeDescription description,
			ENamedElement eNamedElement) {
		EList<FeatureChange> changes = description.getObjectChanges().get(
				eNamedElement);
		EList<EObject> addedObjects = description.getObjectsToAttach();
		EList<EObject> removedObjects = description.getObjectsToDetach();
		gtResources.clear();
		getGTResources(eNamedElement.eResource());
		if (changes != null)
			for (FeatureChange change : changes) {
				System.out.println(change + " Name: " + change.getFeatureName()
						+ " Value: " + change.getDataValue());		
			}
		if (addedObjects != null)
			for (EObject addedObj : addedObjects) {
				System.out.println(addedObj);				
			}
		if (removedObjects != null)
			for (EObject removedObj : removedObjects) {
				System.out.println(removedObj);
			}		
		applyToAffectedObjects2(eNamedElement);
		return false;
	}
}