package de.ubt.ai1.modgraph.example.eRefactoring.connector;

import java.io.IOException;
import java.util.Collections;

import javax.swing.JOptionPane;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.change.ChangeDescription;
import org.eclipse.emf.ecore.change.util.ChangeRecorder;
import org.eclipse.emf.ecore.presentation.EcoreEditor;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.gmf.runtime.notation.BasicCompartment;
import org.eclipse.gmf.runtime.notation.Node;
import org.eclipse.gmf.runtime.notation.View;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;

import de.ubt.ai1.modgraph.gt.failure.GTFailure;
import de.ubt.ai1.modgraph.gt.gtEngine.GPBoundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GPUnboundNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GTAttribute;
import de.ubt.ai1.modgraph.gt.gtEngine.GTNode;
import de.ubt.ai1.modgraph.gt.gtEngine.GTThisObject;
import de.ubt.ai1.modgraph.gt.gtEngine.GraphTransformationRule;
import de.ubt.ai1.modgraph.gt.gtEngine.NACObject;
import de.ubt.ai1.modgraph.gt.gtEngine.NACParameter;
import de.ubt.ai1.modgraph.gtEngine.diagram.part.GTDiagramEditorPlugin;
import de.ubt.ai1.modgraph.gtEngine.diagram.providers.GTViewProvider;

public class ReplaceSubclassWithFieldsConnector extends RefactoringConnector {
	private IEditorPart activeEditor;
	private EClass superClass;

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		activeEditor = HandlerUtil.getActiveEditor(event);
		window = HandlerUtil.getActiveWorkbenchWindow(event);

		if (activeEditor instanceof EcoreEditor) {
			EcoreEditor ecoreEditor = (EcoreEditor) activeEditor;
			
			ecoreEditor.doSave(null);
			TreeSelection sel = (TreeSelection) ecoreEditor.getSelection();
			Object obj = sel.getFirstElement();

			if (obj instanceof EClass) {
				superClass = (EClass) obj;

				initRefactoring(superClass.getEPackage().eContents());

				ChangeRecorder recorder = new ChangeRecorder(superClass);
				try {
					refactoring.replaceSubclassWithFields(superClass.getName());
				} catch (Throwable e) {
					// JOptionPane.showMessageDialog(null,
					// "Refactoring failed.",
					// "ERROR!", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
					return null;
				}
				try {
					superClass.getEPackage().eResource()
					.save(Collections.EMPTY_MAP);
				} catch (IOException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
				try {
					ChangeDescription description = recorder.endRecording();
					if (!(description.eContents().isEmpty())) {
						propagateChanges(description, superClass, superClass.eResource());
					}
				} catch (Throwable e) {
					e.printStackTrace();
				}
			}

		}
		return null;

	}

	@SuppressWarnings("rawtypes")
	@Override
	protected void applyToAffectedObjects(ENamedElement eNamedElement) {
		for (Resource res : gtResources) {
			TreeIterator<EObject> iter = res.getAllContents();
			GraphTransformationRule rule = null;
			while (iter.hasNext()) {
				EObject eObject = iter.next();
				if (eObject instanceof GraphTransformationRule<?, ?>) {
					rule = (GraphTransformationRule) eObject;
					propagation.propagateReplaceSubclassesWithFields(
							superClass, rule);
				}
				else if (eObject instanceof Node) {
					Node node = (Node) eObject;
					EObject element = node.getElement();
					if (element instanceof GPUnboundNode
							|| element instanceof GPBoundNode
							|| element instanceof GTThisObject
							|| element instanceof NACObject
							|| element instanceof NACParameter) {
						EObject type = null;
						if (element instanceof GPUnboundNode){
							type = ((GPUnboundNode) element).getType();							
						}
						if (element instanceof GPBoundNode){
							type = ((GPBoundNode) element).getType();							
						}
						if (element instanceof GTThisObject){
							type = ((GTThisObject) element).getType();							
						}
						if (element instanceof NACObject){
							type = ((NACObject) element).getType();							
						}
						if (element instanceof NACParameter){
							type = ((NACParameter) element).getType();							
						}
						for (EAttribute e : ((EClass) type).getEAttributes()) {
							String enumName = "_" + superClass.getName() + "Type";
							if (e.getEType() instanceof EEnum && e.getName().equals(enumName)) {
								for(GTAttribute gta : ((GTNode) element).getAttributes()){
									if(((ENamedElement) gta.getExAttribute()).getName().equals(enumName)){
										for (Object o : node.getPersistedChildren()) {
											if (o instanceof BasicCompartment) {
												View view = (View) o;
												GTViewProvider vp = new GTViewProvider();
												vp.createGTAttribute_2002(
														gta,
														view,
														0,
														true,
														GTDiagramEditorPlugin.DIAGRAM_PREFERENCES_HINT);
												break;
											}
										}
									}
								}
							}
						}
					}
				}
			}
			propagation.cleanUpAndSave(null, res);
			
			URI resURI = res.getURI();
			String fragment = resURI.toString();
			String[] resRuleNameArray = fragment.split("/");
			String resRuleName = resRuleNameArray[resRuleNameArray.length-1];			
			int response = JOptionPane.showConfirmDialog(null,
					"Would you like to validate the GT rule " + resRuleName + "?", "Attention",
					JOptionPane.YES_NO_OPTION);
			if(response == 0){
				try{
					boolean result = callValidation(res);
					if(result == false){
						JOptionPane.showMessageDialog(null,
								"Errors exist in the rule " + resRuleName + "\nPlease check it manually.", "Validation failed!",
								JOptionPane.ERROR_MESSAGE);
					}
				} catch (Throwable e){
					e.printStackTrace();
				}
			}
		}
	}
}